/* OmniVision OV2610 Camera Chip Support Code
 *
 * Copyright (c) 1999-2003 Mark McClelland <mark@alpha.dyndns.org>
 * http://alpha.dyndns.org/ov511/
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */

/* This file is not a module yet */
#define __NO_VERSION__

#include <linux/config.h>
#include <linux/version.h>
#include <linux/slab.h>
#include "ovcamchip_priv.h"

/* Registers */
#define REG_GAIN		0x00	/* gain [5:0] */
#define REG_BLUE		0x01	/* blue gain MSB */
#define REG_RED			0x02	/* red gain MSB */
#define REG_COMA		0x03	/* [0:1] Red gain LSB, [2:3] Blue gain LSB, [4:7] AWB update threshold */
#define	REG_COMB		0x04	/* [0:2] AEC [4:5] AWB speed [6:7] AWB steps */
#define REG_BAVG		0x05	/* B-channel average */
#define REG_GbAVG		0x06	/* G-channel average for B pixel lines */
#define REG_GrAVG		0x07	/* G-channel average for R pixel lines */
#define	REG_RAVG		0x08	/* R-channel average */
#define REG_COMC		0x09	/* [0:1] Output Drive, [2:3] Sensor sampling, [4] Sleep mode */
#define REG_PIDH		0x0A	/* Product ID MSB */
#define REG_PIDL		0x0B	/* Product ID LSB */
#define REG_COMD		0x0C	/* Common D */
#define REG_AEC			0x10	/* Auto Exposure Control MSB [10:3] */
#define REG_CLKRC		0x11	/* Clock Rate [0:5] Divider [6] Master/slave [7] PLL on/off */
#define REG_COMH		0x12	/* [0] ADC [1] Clr Bars [2] window [3] master/slave [6] UXGA/SVGA [7] reset */
#define REG_COMI		0x13	/* [0] AEC [1] AWB [2] AGC [4:5] BandFilter [6:7] AEC speed */
#define REG_VSTRT		0x19	/* Vertical windowing line MSB start point*/
#define REG_VEND		0x1A	/* Vertical windowing line MSB end point */
#define REG_COML		0x2A	/* Common L */
#define REG_FRARL		0x2B	/* frame rate */
#define REG_ADDVSL		0x2D	/* V-sync pulse width LSB */
#define REG_ADDVSH		0x2E	/* V-sync pulse width MSB */
#define REG_YAVG		0x2F
#define REG_COMM		0x32


struct ov2610 {
	int auto_wb;
	int auto_exp;
	int auto_gain;
	int bandfilt;
	int resolution;
};

/* Settings for OV2610 camera chip */
static struct ovcamchip_regvals regvals_init_2610[] = {
	{ 0x12, 0x80 }, /* reset */
	{ 0xff, 0xff },	/* END MARKER */
};


# define	OV2610_DFL_BRIGHT	0x7F	/* Not really settable */
# define	OV2610_DFL_SAT		0x7F	/* Not really settable */
# define	OV2610_DFL_AUTO_GAIN	0x01	/* Enabled by default */
# define	OV2610_DFL_GAIN		0x00	/* Gain = 1 by default */
# define	OV2610_DFL_AUTO_EXP	0x01	/* Enabled by default */
# define	OV2610_DFL_EXP		0x43	/* Exposure = ?? by default (actual value: 0x43 << 3) */

/* Returns index into the specified look-up table, with 'n' elements, for which
 * the value is greater than or equal to "val". If a match isn't found, (n-1)
 * is returned. The entries in the table must be in ascending order. */
static inline int
ov2610_lut_find(unsigned char lut[], int n, unsigned char val)
{
	int i = 0;

	while (lut[i] < val && i < n)
		i++;

	return i;
}

#if defined(HAVE_V4L2)
static void
ov2610_init_controls(struct ovcamchip *ov)
{

# if 0
	ov->qc_brightness.maximum = 0xff;
	ov->qc_brightness.default_value = OV2610_DFL_BRIGHT;
# else
	// Implement later
	ov->qc_brightness.flags |= V4L2_CTRL_FLAG_DISABLED;
# endif
	ov->qc_contrast.flags |= V4L2_CTRL_FLAG_DISABLED;

# if 0
	ov->qc_saturation.maximum = 0xff;
	ov->qc_saturation.default_value = OV2610_DFL_SAT;
# else
	// Implement later
	ov->qc_saturation.flags |= V4L2_CTRL_FLAG_DISABLED;
# endif

	// Implement later
	ov->qc_red_balance.flags |= V4L2_CTRL_FLAG_DISABLED;

	// Implement later
	ov->qc_blue_balance.flags |= V4L2_CTRL_FLAG_DISABLED;

	ov->qc_agc.default_value = OV2610_DFL_AUTO_GAIN;

	ov->qc_gain.maximum = 0x3f;
	ov->qc_gain.default_value = OV2610_DFL_GAIN;

	if (OV2610_DFL_AUTO_GAIN)
		ov->qc_gain.flags |= V4L2_CTRL_FLAG_GRABBED;

	ov->qc_aec.default_value = OV2610_DFL_AUTO_EXP;
	
	ov->qc_exposure.maximum = 0xff;	/* KLUDGE: we've actually got 11 bits available for this */
	ov->qc_exposure.default_value = OV2610_DFL_EXP;

	if (OV2610_DFL_AUTO_EXP)
		ov->qc_exposure.flags |= V4L2_CTRL_FLAG_GRABBED;

	// Implement later
	ov->qc_awb.flags |= V4L2_CTRL_FLAG_DISABLED;
	return;
}
#endif

/* This initializes the OV2610 camera chip and relevant variables. */
static int
ov2610_init(struct i2c_client *c)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov2610 *s;
	int rc;

	PDEBUG(4, "entered");

	rc = ov_write_regvals(c, regvals_init_2610);

	if (rc < 0)
		return rc;

	ov->spriv = s = kmalloc(sizeof *s, GFP_KERNEL);
	if (!s)
		return -ENOMEM;
	memset(s, 0, sizeof *s);

	s->auto_exp = OV2610_DFL_AUTO_EXP;
	s->auto_gain = OV2610_DFL_AUTO_GAIN;

#if defined(HAVE_V4L2)
	ov2610_init_controls(ov);
#endif
	return 0;
}

static int
ov2610_free(struct i2c_client *c)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);

	kfree(ov->spriv);
	return 0;
}

static int
ov2610_set_v4l1_control(struct i2c_client *c, struct ovcamchip_control *ctl)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov2610 *s = ov->spriv;
	int rc;
	int v = ctl->value;

	switch (ctl->id) {
	case OVCAMCHIP_CID_CONT:
		/* No such thing for RGB chip */
		rc = 0;
		break;

	case OVCAMCHIP_CID_BRIGHT:
		/* cant do it */
		rc = 0;
		break;
		
	case OVCAMCHIP_CID_SAT:
		/* No such thing for RGB chip */
		rc = 0;
		break;
		
	case OVCAMCHIP_CID_EXP:
		if (!s->auto_exp)
			/* KLUDGE: we've got 11 bits of control, but are only using the top 8 */
			/* (the bottom 3 are in REG_COMB if you're really keen to do it right) */
			rc = ov_write(c, REG_AEC, v);
		else
			rc = -EBUSY;
		break;
		
	case OVCAMCHIP_CID_FREQ:
		/* cant do it */
		rc = 0;
		break;

	case OVCAMCHIP_CID_BANDFILT:
		rc = ov_write_mask(c, REG_COMI, v?0x20:0x00, 0x20);
		s->bandfilt = v;
		break;
		
	case OVCAMCHIP_CID_AUTOBRIGHT:
		/* cant do it */
		rc = 0;
		break;
		
	case OVCAMCHIP_CID_AUTOEXP:
		rc = ov_write_mask(c, REG_COMI, v?0x01:0x00, 0x01);
		s->auto_exp = v;
		break;

	case OVCAMCHIP_CID_BACKLIGHT:
		/* cant do it */
		rc = 0;
		break;

	case OVCAMCHIP_CID_MIRROR:
		/* cant do it */
		rc = 0;
		break;
		
	default:
		PDEBUG(2, "control not supported: %d", ctl->id);
		return -EPERM;
	}

	PDEBUG(3, "id=%d, arg=%d, rc=%d", ctl->id, v, rc);
	return rc;
}

static int
ov2610_get_v4l1_control(struct i2c_client *c, struct ovcamchip_control *ctl)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov2610 *s = ov->spriv;
	int rc = 0;
	unsigned char val = 0;

	switch (ctl->id) {
	case OVCAMCHIP_CID_CONT:
		/* can't do it */
		return -EINVAL;
		
	case OVCAMCHIP_CID_BRIGHT:
		/* can't do it */
		return -EINVAL;
		
	case OVCAMCHIP_CID_SAT:
		/* can't do it */
		return -EINVAL;
		
	case OVCAMCHIP_CID_EXP:
		rc = ov_read(c, REG_AEC, &val);
		ctl->value = val;
		break;
		
	case OVCAMCHIP_CID_BANDFILT:
		ctl->value = s->bandfilt;
		break;
		
	case OVCAMCHIP_CID_AUTOBRIGHT:
		/* can't do it */
		return -EINVAL;
		
	case OVCAMCHIP_CID_AUTOEXP:
		ctl->value = s->auto_exp;
		break;
		
	case OVCAMCHIP_CID_BACKLIGHT:
		/* can't do it */
		return -EINVAL;
		
	case OVCAMCHIP_CID_MIRROR:
		/* can't do it */
		return -EINVAL;
		
	default:
		PDEBUG(2, "control not supported: %d", ctl->id);
		return -EPERM;
	}

	PDEBUG(3, "id=%d, arg=%d, rc=%d", ctl->id, ctl->value, rc);
	return rc;
}

#if defined (HAVE_V4L2)
static int
ov2610_set_v4l2_control(struct i2c_client *c, struct v4l2_control *ctl)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov2610 *s = ov->spriv;
	struct v4l2_queryctrl *qc = ovcamchip_get_qc(ov, ctl->id);
	int rc = 0;
	__s32 v = ctl->value;

	if (qc) {
		if (qc->flags & V4L2_CTRL_FLAG_GRABBED) {
			rc = -EBUSY;
			goto out;
		}
	} else {
		rc = -EINVAL;
		goto out;
	}

	switch (ctl->id) {
	case V4L2_CID_BRIGHTNESS:
		rc = -EINVAL;
		break;
		
	case V4L2_CID_SATURATION:
		rc = -EINVAL;
		break;
		
	case V4L2_CID_AUTOGAIN:
		rc = ov_write_mask(c, REG_COMI, v?0x04:0x00, 0x04);
		
		if (rc >= 0) {
			s->auto_gain = v;
			ov->qc_gain.flags &= ~V4L2_CTRL_FLAG_GRABBED;
			ov->qc_gain.flags |= v?V4L2_CTRL_FLAG_GRABBED:0;
		}
		break;
		
	case V4L2_CID_GAIN:
		/* FIXME: This setting may not be linear. It should use a formula
		 * or look-up table for converting reg. val to/from percent.
		 * That may not work well within the constraints of V4L2 tho.
		 */
		if (!s->auto_gain)
			rc = ov_write(c, REG_GAIN, v);
		else
			rc = -EINVAL;
		break;
		
	case V4L2_CID_EXPOSURE:
		if (!s->auto_exp)
			rc = ov_write(c, REG_AEC, v);
		else
			rc = -EINVAL;
		break;
		
	case OVCAMCHIP_V4L2_CID_AEC:
		rc = ov_write_mask(c, REG_COMI, v?0x00:0x01, 0x01);
		if (rc >= 0) {
			s->auto_exp = v;
			ov->qc_exposure.flags &= ~V4L2_CTRL_FLAG_GRABBED;
			ov->qc_exposure.flags |= v?V4L2_CTRL_FLAG_GRABBED:0;
		}
		break;
	default:
		rc = -EINVAL;
	}

out:
	PDEBUG(3, "id=%d, val=%d, rc=%d", ctl->id, ctl->value, rc);
	return rc;
}

static int
ov2610_get_v4l2_control(struct i2c_client *c, struct v4l2_control *ctl)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	struct ov2610 *s = ov->spriv;
	int rc = 0;
	unsigned char val = 0;

	switch (ctl->id) {
	case V4L2_CID_BRIGHTNESS:
		rc = -EINVAL;
		break;
	case V4L2_CID_SATURATION:
		rc = -EINVAL;
		break;
	case V4L2_CID_AUTOGAIN:
		ctl->value = s->auto_gain;
		break;
	case V4L2_CID_GAIN:
		rc = ov_read(c, REG_GAIN, &val);
		ctl->value = val;
		break;
	case V4L2_CID_EXPOSURE:
		rc = ov_read(c, REG_AEC, &val);
		ctl->value = val;
		break;
	case OVCAMCHIP_V4L2_CID_AEC:
		ctl->value = s->auto_exp;
		break;
	default:
		rc = -EINVAL;
	}

	PDEBUG(3, "id=%d, val=%d, rc=%d", ctl->id, ctl->value, rc);
	return (rc<0)?rc:0;
}
#endif

# if 0
static int
ov2610_mode_init(struct i2c_client *c, struct ovcamchip_window *win)
{
	struct ovcamchip *ov = i2c_get_clientdata(c);
	int qvga = win->quarter;
	
	/******** QVGA-specific regs ********/
	ov_write_mask(c, 0x14, qvga?0x20:0x00, 0x20);
	ov_write_mask(c, 0x28, qvga?0x00:0x20, 0x20);
	ov_write(c, 0x24, qvga?0x20:0x3a);
	ov_write(c, 0x25, qvga?0x30:0x60);
	ov_write_mask(c, 0x2d, qvga?0x40:0x00, 0x40);
	if (!ov->mono)
		ov_write_mask(c, 0x67, qvga?0xf0:0x90, 0xf0);
	ov_write_mask(c, 0x74, qvga?0x20:0x00, 0x20);

	/******** Clock programming ********/

	ov_write(c, 0x11, win->clockdiv);
	return 0;
}
# endif

static int
ov2610_set_window(struct i2c_client *c, struct ovcamchip_window *win)
{
# if 0
	int ret, hwscale, vwscale;
	
	info("Asked to window (%d,%d) %dx%d", win->x, win->y, win->width, win->height);

	ret = ov2610_mode_init(c, win);
	if (ret < 0)
		return ret;

	if (win->quarter) {
		hwscale = 1;
		vwscale = 0;
	} else {
		hwscale = 2;
		vwscale = 1;
	}

	ov_write(c, 0x17, HWSBASE + (win->x >> hwscale));
	ov_write(c, 0x18, HWEBASE + ((win->x + win->width) >> hwscale));
	ov_write(c, 0x19, VWSBASE + (win->y >> vwscale));
	ov_write(c, 0x1a, VWEBASE + ((win->y + win->height) >> vwscale));
# endif
	return 0;
}

static int
ov2610_command(struct i2c_client *c, unsigned int cmd, void *arg)
{
	switch (cmd) {
	case OVCAMCHIP_CMD_S_CTRL:
		return ov2610_set_v4l1_control(c, arg);
	case OVCAMCHIP_CMD_G_CTRL:
		return ov2610_get_v4l1_control(c, arg);
#if defined (HAVE_V4L2)
	case VIDIOC_S_CTRL:
		return ov2610_set_v4l2_control(c, arg);
	case VIDIOC_G_CTRL:
		return ov2610_get_v4l2_control(c, arg);
#endif
	case OVCAMCHIP_CMD_S_MODE:
		return ov2610_set_window(c, arg);
	default:
		PDEBUG(2, "command not supported: 0x%X", cmd);
		return -ENOIOCTLCMD;
	}
}

struct ovcamchip_ops ov2610_ops = {
	.init    =	ov2610_init,
	.free    =	ov2610_free,
	.command =	ov2610_command,
};
