/* OmniVision/Cypress FX2 Camera-to-USB Bridge Driver
 *
 * Copyright (c) 1999-2004 Mark W. McClelland, David Brownell et al
 * OV2610 fixes by Steve Hosgood <steve@caederus.com> Jul-Nov 2004
 * Many improvements by Bret Wallach <bwallac1@san.rr.com>
 * Color fixes by by Orion Sky Lawlor <olawlor@acm.org> (2/26/2000)
 * OV7620 fixes by Charl P. Botha <cpbotha@ieee.org>
 * Changes by Claudio Matsuoka <claudio@conectiva.com>
 * Kernel I2C interface improvements by Kysti Mlkki
 * URB error messages from pwc driver by Nemosoft
 * Memory management (rvmalloc) code from bttv driver, by Gerd Knorr and others
 *
 * Based on the Linux CPiA driver written by Peter Pregler,
 * Scott J. Bertin and Johannes Erdfelt.
 * 
 * Please see the file: linux/Documentation/usb/ov511.txt 
 * and the website at:  http://alpha.dyndns.org/ov511
 * for more info.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */
 
#include <linux/config.h>
#include <linux/version.h>

/* 2.6 Doesn't support /proc/video, but this is still defined somewhere */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#  undef CONFIG_VIDEO_PROC_FS
#endif

#include <linux/module.h>
#include <linux/init.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#if defined(CONFIG_VIDEO_PROC_FS)
#  include <linux/fs.h>
#  include <linux/proc_fs.h>
#  include <asm/io.h>
#endif
#include <linux/ctype.h>
#include <linux/pagemap.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 68)
#  include <linux/wrapper.h>
#endif
#include <linux/mm.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
#  include <linux/device.h>
#endif

#define MODULE_DEBUG
#include "ovfx2.h"

# include "bayer.h"
# include "bayer_bgbg_grgr.h"	/* KLUDGE - only the sensor module ought to be working at this level */
# include "id.h"

/* Driver will compile with older kernels, but will oops on open() */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 19)
#  error "Kernel 2.4.19 is the minimum for this driver"
#endif

/*
 * Version Information
 */
#define DRIVER_VERSION "v2.27"
#define EMAIL "mark@alpha.dyndns.org"
#define DRIVER_AUTHOR "Mark McClelland <mark@alpha.dyndns.org> & Bret Wallach \
	& Orion Sky Lawlor <olawlor@acm.org> & Kevin Moore & Charl P. Botha \
	<cpbotha@ieee.org> & Claudio Matsuoka <claudio@conectiva.com> & Steve \
	Hosgood <steve@caederus.com>"
#define DRIVER_DESC "ovfx2 USB Camera Driver"

//#define STOPSTREAM 0x1D	/* stop/pause stream command (0x1D, 0x15, 0x01 work) */
//#define STREAM	0x1F	/* start stream command (0x1F, 0x1B, 0x17, 0x0F work) */

#define STOPSTREAM 0x00	/* FX2's 0x0F register */
#define STARTSTREAM 0x02
#define STREAMMASK 0x02
#define STOPUNK 0x00	/* FX2's 0x0F register */
#define STARTUNK 0x04
#define UNKMASK 0x04
#define STOPSVGA 0x00	/* FX2's 0x0F register */
#define STARTSVGA 0x80
#define SVGAMASK 0x80

#define FIFO16	0x00	/* FX2's 0x02 register (0x00 runs the FIFO in 16-bit mode, 0x01, 0x02, 0x04, 0x08 do 8-bit) */
#define FIFO8	0x01

#define GAIN1	0x00
#define GAIN1_5	0x08
#define GAIN2	0x10
#define GAIN3	0x18
#define GAIN4	0x30
#define GAIN5	0x34
#define GAIN6	0x38
#define GAIN7	0x3C
#define GAIN8	0x3F	/* actually gain of 7.75 */

#define UXGA_SPEED	0x80		/* speed: 0x80 = 10fps, 0x81 = 5fps ( >= 0x82 doesn't work) */
#define UXGA_GAIN	GAIN2
#define SVGA_SPEED	0x80		/* speed: 0x80 = 40fps, 0x81 = 20fps, 0x82 = 13.3fps, 0x83 = 10fps ( >= 0x84 doesn't work) */
#define SVGA_GAIN	GAIN4		/* compensates for quad frame-rate */

/* Pixel count * 4 bytes per pixel (for BGR24) */
#define MAX_FRAME_SIZE(w, h) ((w) * (h) * 4)

#define MAX_DATA_SIZE(w, h) (MAX_FRAME_SIZE(w, h) + sizeof(struct timeval))

# ifdef HARDCODED_SENSOR
/* Max size * 2 bytes per pixel average + safety margin */
#define MAX_RAW_DATA_SIZE(w, h) ((w) * (h) * 2)
# else
/* Max individual raw frame size: 1 bytes per pixel */
#define MAX_RAW_DATA_SIZE(w, h) ((w) * (h))
# endif

#define FATAL_ERROR(rc) ((rc) < 0 && (rc) != -EPERM)

/**********************************************************************
 * Module Parameters
 * (See ov511.txt for detailed descriptions of these)
 **********************************************************************/

/* These variables (and all static globals) default to zero */
static int autobright		= 1;
static int autoexp		= 1;
#ifdef MODULE_DEBUG
static int debug;
#endif
static int dumppix;
#ifdef MODULE_DEBUG
static int dump_bridge;
#endif
static int lightfreq;
static int bandingfilter;
static int clockdiv		= -1;
# ifdef HARDCODED_SENSOR
static int framedrop		= -1;
# endif
static int fastset;
static int force_palette;
static int backlight;
static int mirror;

MODULE_PARM(autobright, "i");
MODULE_PARM_DESC(autobright, "Sensor automatically changes brightness");
MODULE_PARM(autoexp, "i");
MODULE_PARM_DESC(autoexp, "Sensor automatically changes exposure");
#ifdef MODULE_DEBUG
MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug,
  "Debug level: 0=none, 1=inits, 2=warning, 3=config, 4=functions, 5=max");
#endif
MODULE_PARM(dumppix, "i");
MODULE_PARM_DESC(dumppix, "Dump raw pixel data");
#ifdef MODULE_DEBUG
MODULE_PARM(dump_bridge, "i");
MODULE_PARM_DESC(dump_bridge, "Dump the bridge registers");
#endif
MODULE_PARM(lightfreq, "i");
MODULE_PARM_DESC(lightfreq,
  "Light frequency. Set to 50 or 60 Hz, or zero for default settings");
MODULE_PARM(bandingfilter, "i");
MODULE_PARM_DESC(bandingfilter,
  "Enable banding filter (to reduce effects of fluorescent lighting)");
MODULE_PARM(clockdiv, "i");
MODULE_PARM_DESC(clockdiv, "Force pixel clock divisor to a specific value");
# ifdef HARDCODED_SENSOR
MODULE_PARM(framedrop, "i");
MODULE_PARM_DESC(framedrop, "Force a specific frame drop register setting");
# endif
MODULE_PARM(fastset, "i");
MODULE_PARM_DESC(fastset, "Allows picture settings to take effect immediately");
MODULE_PARM(force_palette, "i");
MODULE_PARM_DESC(force_palette, "Force the palette to a specific value");
MODULE_PARM(backlight, "i");
MODULE_PARM_DESC(backlight, "For objects that are lit from behind");
MODULE_PARM(mirror, "i");
MODULE_PARM_DESC(mirror, "Reverse image horizontally");

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
#if defined(MODULE_LICENSE)	/* Introduced in ~2.4.10 */
MODULE_LICENSE("GPL");
#endif

/* mention a default FourCC packer */
static unsigned char *Cam2RGB24( unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash );


/**********************************************************************
 * Miscellaneous Globals
 **********************************************************************/

/* Prevents double-free of ov struct */
static struct semaphore ov_free_lock;

# if 0
/*
 * Omnivision provided a reference design and firmware connecting an FX2
 * to many of their I2C-aware sensor arrays.  Different products using
 * this design could have different sensors.
 */
static struct usb_device_id device_table [] = {
	/*
	 * Orange Micro "iBOT2" uses an ov7620.
	 *
	 * HIGH SPEED:
	 *
	 * T:  Bus=04 Lev=01 Prnt=01 Port=00 Cnt=01 Dev#=  2 Spd=480 MxCh= 0
	 * D:  Ver= 2.00 Cls=00(>ifc ) Sub=00 Prot=00 MxPS=64 #Cfgs=  1
	 * P:  Vendor=0b62 ProdID=0059 Rev= 1.00
	 * S:  Manufacturer=Orange Micro
	 * S:  Product=iBOT2 Camera
	 * C:* #Ifs= 1 Cfg#= 1 Atr=80 MxPwr=400mA
	 * I:  If#= 0 Alt= 0 #EPs= 2 Cls=ff(vend.) Sub=00 Prot=00 Driver=(none)
	 * E:  Ad=81(I) Atr=03(Int.) MxPS=   4 Ivl=  4ms
	 * E:  Ad=82(I) Atr=02(Bulk) MxPS= 512 Ivl=  1ms
	 *
	 * FULL SPEED:
	 *
	 * T:  Bus=02 Lev=01 Prnt=01 Port=00 Cnt=01 Dev#=  3 Spd=12  MxCh= 0
	 * D:  Ver= 2.00 Cls=00(>ifc ) Sub=00 Prot=00 MxPS=64 #Cfgs=  1
	 * P:  Vendor=0b62 ProdID=0059 Rev= 1.00
	 * S:  Manufacturer=Orange Micro
	 * S:  Product=iBOT2 Camera
	 * C:* #Ifs= 1 Cfg#= 1 Atr=80 MxPwr=400mA
	 * I:  If#= 0 Alt= 0 #EPs= 2 Cls=ff(vend.) Sub=00 Prot=00 Driver=(none)
	 * E:  Ad=81(I) Atr=03(Int.) MxPS=   4 Ivl=  1ms
	 * E:  Ad=82(I) Atr=02(Bulk) MxPS=  64 Ivl=  1ms
	 *
	 */
	{ USB_DEVICE (VEND_ORANGE_MICRO, PROD_IBOT2),
		.driver_info = (unsigned long) "Orange Micro iBOT2",
		},
	{ USB_DEVICE (VEND_OMNIVISION, PROD_TEST),
		.driver_info = (unsigned long) "OV PC Cam",
		},
	{ USB_DEVICE (VEND_XLI, PROD_XLI_2M),
		.driver_info = (unsigned long) "XLI 2.1 MegaPixel Camera",
		},

	{ }
};

MODULE_DEVICE_TABLE (usb, device_table);
# endif

/**********************************************************************
 * Symbolic Names
 **********************************************************************/

/* Video4Linux1 Palettes */
static struct symbolic_list v4l1_plist[] = {
	{ VIDEO_PALETTE_GREY,	"GREY" },
	{ VIDEO_PALETTE_HI240,	"HI240" },
	{ VIDEO_PALETTE_RGB565,	"RGB565" },
	{ VIDEO_PALETTE_RGB24,	"RGB24" },
	{ VIDEO_PALETTE_RGB32,	"RGB32" },
	{ VIDEO_PALETTE_RGB555,	"RGB555" },
	{ VIDEO_PALETTE_YUV422,	"YUV422" },
	{ VIDEO_PALETTE_YUYV,	"YUYV" },
	{ VIDEO_PALETTE_UYVY,	"UYVY" },
	{ VIDEO_PALETTE_YUV420,	"YUV420" },
	{ VIDEO_PALETTE_YUV411,	"YUV411" },
	{ VIDEO_PALETTE_RAW,	"RAW" },
	{ VIDEO_PALETTE_YUV422P,"YUV422P" },
	{ VIDEO_PALETTE_YUV411P,"YUV411P" },
	{ VIDEO_PALETTE_YUV420P,"YUV420P" },
	{ VIDEO_PALETTE_YUV410P,"YUV410P" },
	{ -1, NULL }
};

#if defined(CONFIG_VIDEO_PROC_FS) \
 || LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
# if 0
static struct symbolic_list senlist[] = {
	{ CC_OV76BE,	"OV76BE" },
	{ CC_OV7610,	"OV7610" },
	{ CC_OV7620,	"OV7620" },
	{ CC_OV7620AE,	"OV7620AE" },
	{ CC_OV6620,	"OV6620" },
	{ CC_OV6630,	"OV6630" },
	{ CC_OV6630AE,	"OV6630AE" },
	{ CC_OV6630AF,	"OV6630AF" },
	{ CC_OV2610,	"OV2610" },
	{ -1, NULL }
};
#endif
#endif

/* URB error codes: */
static struct symbolic_list urb_errlist[] = {
	{ -ENOSR,	"Buffer error (overrun)" },
	{ -EPIPE,	"Stalled (device not responding)" },
	{ -EOVERFLOW,	"Babble (bad cable?)" },
	{ -EPROTO,	"Bit-stuff error (bad cable?)" },
	{ -EILSEQ,	"CRC/Timeout" },
	{ -ETIMEDOUT,	"NAK (device does not respond)" },
	{ -1, NULL }
};

#ifdef MODULE_DEBUG
static const char *v4l1_ioctls[] = {
	"?", "GCAP", "GCHAN", "SCHAN", "GTUNER", "STUNER", "GPICT", "SPICT",
	"CAPTURE", "GWIN", "SWIN", "GFBUF", "SFBUF", "KEY", "GFREQ",
	"SFREQ", "GAUDIO", "SAUDIO", "SYNC", "MCAPTURE", "GMBUF", "GUNIT",
	"GCAPTURE", "SCAPTURE", "SPLAYMODE", "SWRITEMODE", "GPLAYINFO",
	"SMICROCODE", "GVBIFMT", "SVBIFMT" };
#define NUM_V4L1_IOCTLS (sizeof(v4l1_ioctls)/sizeof(char*))
#endif
/**********************************************************************
 * Prototypes
 **********************************************************************/

static int ov7xx0_configure(struct usb_ovfx2 *);
static int ov2xx0_configure(struct usb_ovfx2 *);

/**********************************************************************
 * Memory management
 **********************************************************************/

/* Here we want the physical address of the memory.
 * This is used when initializing the contents of the area.
 */
static inline unsigned long
kvirt_to_pa(unsigned long adr)
{
	unsigned long kva, ret;

	kva = (unsigned long) page_address(vmalloc_to_page((void *)adr));
	kva |= adr & (PAGE_SIZE-1); /* restore the offset */
	ret = __pa(kva);
	return ret;
}

static void *
rvmalloc(unsigned long size)
{
	void *mem;
	unsigned long adr;

	size = PAGE_ALIGN(size);
	mem = vmalloc_32(size);
	if (!mem)
		return NULL;

	memset(mem, 0, size); /* Clear the ram out, no junk to the user */
	adr = (unsigned long) mem;
	while (size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		SetPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_reserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}

	return mem;
}

static void
rvfree(void *mem, unsigned long size)
{
	unsigned long adr;

	if (!mem)
		return;

	adr = (unsigned long) mem;
	while ((long) size > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 68)
		ClearPageReserved(vmalloc_to_page((void *)adr));
#else
		mem_map_unreserve(vmalloc_to_page((void *)adr));
#endif
		adr += PAGE_SIZE;
		size -= PAGE_SIZE;
	}
	vfree(mem);
}

/**********************************************************************
 * /proc interface
 * Based on the CPiA driver version 0.7.4 -claudio
 **********************************************************************/

#if defined(CONFIG_VIDEO_PROC_FS)
#error "CONFIG_VIDEO_PROC_FS probably doesn't work yet"
static struct proc_dir_entry *ovfx2_proc_entry = NULL;
extern struct proc_dir_entry *video_proc_entry;

/* Prototypes */
static int sensor_get_picture(struct usb_ovfx2 *, struct video_picture *);
static int sensor_get_control(struct usb_ovfx2 *, int, int *);

#define YES_NO(x) ((x) ? "yes" : "no")

/* /proc/video/ovfx2/<minor#>/info */
static int
ovfx2_read_proc_info(char *page, char **start, off_t off, int count, int *eof,
		     void *data)
{
	char *out = page;
	int i, len;
	struct usb_ovfx2 *ov = data;
	struct video_picture p;
	int exp = 0;

	if (!ov || !ov->dev)
		return -ENODEV;

	sensor_get_picture(ov, &p);
	sensor_get_control(ov, OVCAMCHIP_CID_EXP, &exp);

	/* IMPORTANT: This output MUST be kept under PAGE_SIZE
	 *            or we need to get more sophisticated. */

	out += sprintf(out, "driver_version  : %s\n", DRIVER_VERSION);
	out += sprintf(out, "model           : %s\n", ov->desc);
	out += sprintf(out, "streaming       : %s\n", YES_NO(ov->streaming));
	out += sprintf(out, "grabbing        : %s\n", YES_NO(ov->grabbing));
	out += sprintf(out, "subcapture      : %s\n", YES_NO(ov->sub_flag));
	out += sprintf(out, "sub_size        : %d %d %d %d\n",
		       ov->subx, ov->suby, ov->subw, ov->subh);
	out += sprintf(out, "brightness      : %d\n", p.brightness >> 8);
	out += sprintf(out, "colour          : %d\n", p.colour >> 8);
	out += sprintf(out, "contrast        : %d\n", p.contrast >> 8);
	out += sprintf(out, "hue             : %d\n", p.hue >> 8);
	out += sprintf(out, "exposure        : %d\n", exp);
	out += sprintf(out, "num_frames      : %d\n", OVFX2_NUMFRAMES);
	for (i = 0; i < OVFX2_NUMFRAMES; i++) {
		out += sprintf(out, "frame           : %d\n", i);
		out += sprintf(out, "  size          : %d %d\n",
			       ov->frame[i].width, ov->frame[i].height);
		out += sprintf(out, "  format        : %s\n",
			       symbolic(v4l1_plist, ov->frame[i].format));
	}
	out += sprintf(out, "sensor          : %s\n",
		       symbolic(senlist, ov->sensor));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 20)
	out += sprintf(out, "topology        : %s\n", ov->usb_path);
#else
	out += sprintf(out, "usb_bus         : %d\n", ov->dev->bus->busnum);
	out += sprintf(out, "usb_device      : %d\n", ov->dev->devnum);
#endif
	out += sprintf(out, "i2c_adap_id     : %d\n",
	               i2c_adapter_id(&ov->i2c_adap));

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

static void
create_proc_ovfx2_cam(struct usb_ovfx2 *ov)
{
	char dirname[10];

	if (!ovfx2_proc_entry || !ov) {
		info("No /proc entry to add to");
		return;
	}else
		info("Adding a /proc entry");

	/* Create per-device directory */
	snprintf(dirname, 10, "%d", ov->minor);
	ov->proc_devdir = create_proc_entry(dirname, S_IFDIR, ovfx2_proc_entry);
	if (!ov->proc_devdir)
		return;
	ov->proc_devdir->owner = THIS_MODULE;

	/* Create "info" entry (human readable device information) */
	ov->proc_info = create_proc_read_entry("info", S_IFREG|S_IRUGO|S_IWUSR,
		ov->proc_devdir, ovfx2_read_proc_info, ov);
	if (!ov->proc_info)
		return;
	ov->proc_info->owner = THIS_MODULE;
}

static void
destroy_proc_ovfx2_cam(struct usb_ovfx2 *ov)
{
	char dirname[10];

	if (!ov || !ov->proc_devdir)
		return;

	snprintf(dirname, 10, "%d", ov->minor);

	/* Destroy "info" entry */
	if (ov->proc_info) {
		remove_proc_entry("info", ov->proc_devdir);
		ov->proc_info = NULL;
	}

	/* Destroy per-device directory */
	remove_proc_entry(dirname, ovfx2_proc_entry);
	ov->proc_devdir = NULL;
}

static void
proc_ovfx2_create(void)
{
	if (video_proc_entry == NULL) {
		err("Error: /proc/video/ does not exist");
		return;
	}

	ovfx2_proc_entry = create_proc_entry("ovfx2", S_IFDIR,
					     video_proc_entry);
	info("Created proc entry");

	if (ovfx2_proc_entry)
		ovfx2_proc_entry->owner = THIS_MODULE;
	else
		err("Unable to create /proc/video/ovfx2");
}

static void
proc_ovfx2_destroy(void)
{
	if (ovfx2_proc_entry == NULL)
		return;

	remove_proc_entry("ovfx2", video_proc_entry);
}
#else
static inline void create_proc_ovfx2_cam(struct usb_ovfx2 *ov) { }
static inline void destroy_proc_ovfx2_cam(struct usb_ovfx2 *ov) { }
static inline void proc_ovfx2_create(void) { }
static inline void proc_ovfx2_destroy(void) { }
#endif /* #ifdef CONFIG_VIDEO_PROC_FS */

/**********************************************************************
 *
 * Register I/O
 *
 **********************************************************************/

/* Write an OVFX2 register */
static int
reg_w(struct usb_ovfx2 *ov, unsigned char reg, unsigned char value)
{
	int rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_device *dev = interface_to_usbdev(ov->intf);
#else
	struct usb_device *dev = ov->dev;
#endif

	PDEBUG(OUTGOING|5, "0x%02X:0x%02X", reg, value);

	/* We don't use cbuf here, but the lock ensures we're done before
	 * disconnect() completes */
	down(&ov->cbuf_lock);

	if (!ov->cbuf) {
		up(&ov->cbuf_lock);
		return -ENODEV;
	}

	rc = usb_control_msg(dev,
			     usb_sndctrlpipe(dev, 0),
			     OVFX2_REQ_REG_WRITE,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			     (__u16)value, (__u16)reg, NULL, 0, HZ);

	up(&ov->cbuf_lock);

	if (rc < 0)
		err("reg write: error %d: %s", rc, symbolic(urb_errlist, rc));

	return rc;
}

/* Read from an OVFX2 register */
/* returns: negative is error, pos or zero is data */
static int
reg_r(struct usb_ovfx2 *ov, unsigned char reg)
{
	int rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_device *dev = interface_to_usbdev(ov->intf);
#else
	struct usb_device *dev = ov->dev;
#endif

	down(&ov->cbuf_lock);

	if (!ov->cbuf) {
		up(&ov->cbuf_lock);
		return -ENODEV;
	}

	rc = usb_control_msg(dev,
			     usb_rcvctrlpipe(dev, 0),
			     OVFX2_REQ_REG_READ,
			     USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			     0, (__u16)reg, &ov->cbuf[0], 1, HZ);

	if (rc < 0) {
		err("reg read: error %d: %s", rc, symbolic(urb_errlist, rc));
	} else {
		rc = ov->cbuf[0];
		PDEBUG(OUTGOING|5, "0x%02X:0x%02X", reg, ov->cbuf[0]);
	}

	up(&ov->cbuf_lock);

	return rc;
}

/*
 * Writes bits at positions specified by mask to an OVFX2 reg. Bits that are in
 * the same position as 1's in "mask" are cleared and set to "value". Bits
 * that are in the same position as 0's in "mask" are preserved, regardless
 * of their respective state in "value".
 */
static int
reg_w_mask(struct usb_ovfx2 *ov,
	   unsigned char reg,
	   unsigned char value,
	   unsigned char mask)
{
	int ret;
	unsigned char oldval, newval;

	if (mask == 0xff) {
		newval = value;
	} else {
		ret = reg_r(ov, reg);
		if (ret < 0)
			return ret;

		oldval = (unsigned char) ret;
		oldval &= (~mask);		/* Clear the masked bits */
		value &= mask;			/* Enforce mask on value */
		newval = oldval | value;	/* Set the desired bits */
	}

	return (reg_w(ov, reg, newval));
}


static int
i2c_r(struct usb_ovfx2 *ov, unsigned char reg )
{
	return i2c_smbus_read_byte_data(&ov->internal_client, reg);
}

static int
i2c_w(struct usb_ovfx2 *ov, unsigned char reg, unsigned char value)
{
	return i2c_smbus_write_byte_data(&ov->internal_client, reg, value);
}

static int
i2c_w_mask(struct usb_ovfx2 *ov,
	   unsigned char reg,
	   unsigned char value,
	   unsigned char mask)
{
	int rc;
	unsigned char oldval, newval;

	if (mask == 0xff) {
		newval = value;
	} else {
		rc = i2c_smbus_read_byte_data(&ov->internal_client, reg);
		if (rc < 0)
			return rc;

		oldval = (unsigned char) rc;
		oldval &= (~mask);		/* Clear the masked bits */
		value &= mask;			/* Enforce mask on value */
		newval = oldval | value;	/* Set the desired bits */
	}

	return i2c_smbus_write_byte_data(&ov->internal_client, reg, newval);
}


/**********************************************************************
 *
 * Low-level I2C I/O functions
 *
 **********************************************************************/

/* NOTE: Do not call this function directly!
 * Sets I2C read and write slave IDs, if they have changed.
 * Returns <0 for error
 */
static int
ovfx2_i2c_adap_set_slave(struct usb_ovfx2 *ov, unsigned char slave)
{
	int rc;

	if (ov->last_slave == slave)
		return 0;

	/* invalidate slave */
	ov->last_slave = 0;

	rc = reg_w(ov, REG_I2C_ADDR, (slave<<1));
	if (rc < 0)
		return rc;

	/* validate slave */
	ov->last_slave = slave;
	return 0;
}

static int
ovfx2_i2c_adap_read_byte_data(struct usb_ovfx2 *ov,
			      unsigned char addr,
			      unsigned char subaddr,
			      unsigned char *val)
{
	int rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_device *dev = interface_to_usbdev(ov->intf);
#else
	struct usb_device *dev = ov->dev;
#endif

	/* Set slave addresses */	
	rc = ovfx2_i2c_adap_set_slave(ov, addr);
	if (rc < 0)
		return rc;

	down(&ov->cbuf_lock);

	if (!ov->cbuf) {
		up(&ov->cbuf_lock);
		return -ENODEV;
	}

	rc = usb_control_msg(dev,
			     usb_rcvctrlpipe(dev, 0),
			     OVFX2_REQ_I2C_READ,
			     USB_DIR_IN | USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			     0, (__u16)subaddr, &ov->cbuf[0], 1, HZ);

	if (rc < 0) {
		PDEBUG(MISC|5, "error %d: %s", rc, symbolic(urb_errlist, rc));
	} else {
		*val = ov->cbuf[0];
		PDEBUG(OUTGOING|5, "(0x%02X) 0x%02X:0x%02X", addr << 1, subaddr, ov->cbuf[0]);
	}

	up(&ov->cbuf_lock);

	return rc;
}

static int
ovfx2_i2c_adap_write_byte_data(struct usb_ovfx2 *ov,
			       unsigned char addr,
			       unsigned char subaddr,
			       unsigned char val)
{
	int rc;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_device *dev = interface_to_usbdev(ov->intf);
#else
	struct usb_device *dev = ov->dev;
#endif

	PDEBUG(OUTGOING|5, "(0x%02X) 0x%02X:0x%02X", addr<<1, subaddr, val);

	/* Set slave addresses */	
	rc = ovfx2_i2c_adap_set_slave(ov, addr);
	if (rc < 0)
		return rc;

	/* We don't use cbuf here, but the lock ensures we're done before
	 * disconnect() completes */
	down(&ov->cbuf_lock);

	if (!ov->cbuf) {
		up(&ov->cbuf_lock);
		return -ENODEV;
	}

	rc = usb_control_msg(dev,
			     usb_sndctrlpipe(dev, 0),
			     OVFX2_REQ_I2C_WRITE,
			     USB_TYPE_VENDOR | USB_RECIP_DEVICE,
			     (__u16)val, (__u16)subaddr, NULL, 0, HZ);

	up(&ov->cbuf_lock);

	if (rc < 0)
		PDEBUG(MISC|5, "error %d: %s", rc, symbolic(urb_errlist, rc));

	return rc;
}

static int
write_regvals(struct usb_ovfx2 *ov, struct regval *regvals)
{
	int rc;

	while (regvals->mask != 0) {
		rc = reg_w_mask(ov, regvals->reg, regvals->val, regvals->mask);
		if (rc < 0)
			return rc;

		regvals++;
	}

	return 0;
}

#ifdef MODULE_DEBUG
static void
dump_reg_range(struct usb_ovfx2 *ov, int reg1, int regn)
{
	int i, rc;

	for (i = reg1; i <= regn; i++) {
		rc = reg_r(ov, i);
		info("OVFX2[0x%02X] = 0x%02X", i, rc);
	}
}

static void
ovfx2_dump_regs(struct usb_ovfx2 *ov)
{
	dump_reg_range(ov, 0x00, 0xff);
}
#endif

/**********************************************************************
 *
 * Kernel I2C Interface
 *
 **********************************************************************/

static int
ovfx2_i2c_validate_addr(struct usb_ovfx2 *ov, u16 addr)
{
	switch (addr) {
	case OV2xx0_SID:
		return 0;
	case OV7xx0_SID:
		return 0;
	default:
		PDEBUG(MISC|4, "Rejected slave ID 0x%04X", addr);
		return -EINVAL;
	}
}

static inline int
sensor_cmd(struct usb_ovfx2 *ov, unsigned int cmd, void *arg)
{
	struct i2c_client *c = ov->sensor_client;

	if (c && c->driver->command)
		return c->driver->command(ov->sensor_client, cmd, arg);
	else
		return -ENODEV;
}

#if 0
static void
call_i2c_clients(struct usb_ovfx2 *ov, unsigned int cmd, void *arg)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	struct i2c_client *client;
	int i;

	for (i = 0; i < I2C_CLIENT_MAX; i++) {
		client = ov->i2c_adap.clients[i];

		/* Sensor_client is not called from here for now, since it
		   has different enumeration for cmd  */
		if (client == ov->sensor_client)
			continue;

		if (client && client->driver->command)
			client->driver->command(client, cmd, arg);
	}
#else
	i2c_clients_command(&ov->i2c_adap, cmd, arg);
#endif
}
#endif

static int
ovfx2_smbus_xfer(struct i2c_adapter *adapter, u16 addr, unsigned short flags,
		 char read_write, u8 command, int size,
		 union i2c_smbus_data *data)
{
	struct usb_ovfx2 *ov = i2c_get_adapdata(adapter);
	int rc = -ENOSYS;

	if (size == I2C_SMBUS_QUICK) {
		PDEBUG(MISC|4, "Got probed at addr 0x%04X", addr);
		rc = ovfx2_i2c_validate_addr(ov, addr);
	} else if (size == I2C_SMBUS_BYTE_DATA) {
		if (read_write == I2C_SMBUS_WRITE) {
 			rc = ovfx2_i2c_adap_write_byte_data(ov, addr, command,
							    data->byte);
		} else if (read_write == I2C_SMBUS_READ) {
			rc = ovfx2_i2c_adap_read_byte_data(ov, addr, command,
							   &data->byte);
		}
	} else {
		warn("Unsupported I2C transfer mode (%d)", size);
		return -EINVAL;
	}

	/* This works around a bug in the I2C core */
	if (rc > 0)
		rc = 0;

	return rc;
}

static u32
ovfx2_i2c_func(struct i2c_adapter *adap)
{
	return I2C_FUNC_SMBUS_QUICK
	     | I2C_FUNC_SMBUS_BYTE_DATA;
}

static int
i2c_attach_inform(struct i2c_client *client)
{
	struct usb_ovfx2 *ov = i2c_get_adapdata(client->adapter);

	int id = client->driver->id;
	info("Attach I2C client addr 0x%X", client->addr);

	if (id == I2C_DRIVERID_OVCAMCHIP) {
		int rc, mono = 0;

		ov->sensor_client = client;

		PDEBUG(MISC|4, "call sensor_cmd OVCAMCHIP_CMD_INITIALIZE");
		rc = sensor_cmd(ov, OVCAMCHIP_CMD_INITIALIZE, &mono);
		if (rc < 0) {
			err("ERROR: Sensor init failed (rc=%d)", rc);
			ov->sensor_client = NULL;
			return rc;
		}

		down(&ov->lock);
		PDEBUG(MISC|4, "call sensor_cmd OVCAMCHIP_Q_SUBTYPE");
		if (sensor_cmd(ov, OVCAMCHIP_CMD_Q_SUBTYPE, &ov->sensor) < 0)
			rc = -EIO;
		else 
		if (client->addr == OV7xx0_SID)
			rc = ov7xx0_configure(ov);
		else
		if (client->addr == OV2xx0_SID)
			rc = ov2xx0_configure(ov);
		else
			rc = -EINVAL;
		up(&ov->lock);

		if (rc) {
			ov->sensor_client = NULL;
			return rc;
		}
	} else	{
		PDEBUG(MISC|4, "Rejected client [%s] with [%s]",
		       i2c_clientname(client), client->driver->name);
		return -1;
	}

	PDEBUG(MISC|4, "i2c attach client [%s] with [%s]",
	       i2c_clientname(client), client->driver->name);

	return 0;
}

static int
i2c_detach_inform(struct i2c_client *client)
{
	struct usb_ovfx2 *ov = i2c_get_adapdata(client->adapter);

	if (ov->sensor_client == client) {
		ov->sensor_client = NULL;
	}

	PDEBUG(MISC|4, "i2c detach [%s]", i2c_clientname(client));

	return 0;
}

static int 
ovfx2_i2c_control(struct i2c_adapter *adapter, unsigned int cmd,
		  unsigned long arg)
{
	return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
static void
ovfx2_i2c_inc_use(struct i2c_adapter *adap)
{
	MOD_INC_USE_COUNT;
}

static void
ovfx2_i2c_dec_use(struct i2c_adapter *adap)
{
	MOD_DEC_USE_COUNT;
}
#endif

static struct i2c_algorithm ovfx2_i2c_algo = {
	.name =			"OVFX2 algorithm",
	.id =			I2C_ALGO_SMBUS,
	.smbus_xfer =		ovfx2_smbus_xfer,
	.algo_control =		ovfx2_i2c_control,
	.functionality =	ovfx2_i2c_func,
};

static struct i2c_adapter i2c_adap_template = {
	I2C_DEVNAME("(unset)"),
	.id =			I2C_ALGO_SMBUS,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 6)
	.class =		I2C_CLASS_CAM_DIGITAL,
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 70)
	.class =		I2C_ADAP_CLASS_CAM_DIGITAL,
#endif
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	.inc_use =		ovfx2_i2c_inc_use,
	.dec_use =		ovfx2_i2c_dec_use,
#else
	.owner =		THIS_MODULE,
#endif
	.client_register =	i2c_attach_inform,
	.client_unregister =	i2c_detach_inform,
};

static int
ovfx2_init_i2c(struct usb_ovfx2 *ov)
{
	memcpy(&ov->i2c_adap, &i2c_adap_template, sizeof(struct i2c_adapter));

	/* Temporary name. We'll set the final one when we know the minor # */
	sprintf(i2c_adapname(&ov->i2c_adap), "OVFX2");
	i2c_set_adapdata(&ov->i2c_adap, ov);

	ov->i2c_adap.algo = &ovfx2_i2c_algo;
	ov->i2c_adap.id |= I2C_HW_SMBUS_OVFX2;

	ov->internal_client.adapter = &ov->i2c_adap;

	PDEBUG(MISC|4, "Registering I2C bus with kernel");

	return i2c_add_adapter(&ov->i2c_adap);
}

/*****************************************************************************/

/* Pause video streaming */
static inline int
ovfx2_pause(struct usb_ovfx2 *ov)
{
	PDEBUG(OUTGOING|4, "pausing stream");
	ov->fx2state = FX2_PAUSED;
	return reg_w_mask(ov, 0x0f, STOPSTREAM, STREAMMASK);
}

/* Resume video streaming if it was stopped. */
static inline int
ovfx2_resume(struct usb_ovfx2 *ov)
{
	if (ov->fx2state == FX2_PAUSED) {
		PDEBUG(OUTGOING|4, "resuming stream");
		ov->fx2state = FX2_STREAMING;
		ov->fsync = -2;		/* skip two frames on resumption */
		return reg_w_mask(ov, 0x0f, STARTSTREAM, STREAMMASK);
	}

	return 0;
}

/* Returns 1 if image streaming needs to be stopped while setting the specified
 * control, and returns 0 if not */
static int
sensor_needs_stop(struct usb_ovfx2 *ov, int cid)
{
	if (!ov->stop_during_set)
		return 0;

	/* FIXME: I don't know whether the FX2 needs to be stopped yet. Don't
	 * do it yet, since it might not be safe to pause the stream while bulk
	 * requests are active */
	return 0;

	switch (cid) {
	case OVCAMCHIP_CID_CONT:
	case OVCAMCHIP_CID_BRIGHT:
	case OVCAMCHIP_CID_SAT:
	case OVCAMCHIP_CID_HUE:
	case OVCAMCHIP_CID_EXP:
		 return 1;
	}

	return 0;
}

static int
sensor_set_control(struct usb_ovfx2 *ov, int cid, int val)
{
	struct ovcamchip_control ctl;
	int rc;

	if (sensor_needs_stop(ov, cid))
		if (ovfx2_pause(ov) < 0)
			return -EIO;

	ctl.id = cid;
	ctl.value = val;

	rc = sensor_cmd(ov, OVCAMCHIP_CMD_S_CTRL, &ctl);

	if (ovfx2_resume(ov) < 0)
		return -EIO;

	return rc;
}

# if 0
static int
sensor_get_control(struct usb_ovfx2 *ov, int cid, int *val)
{
	struct ovcamchip_control ctl;
	int rc;

	ctl.id = cid;

	rc = sensor_cmd(ov, OVCAMCHIP_CMD_G_CTRL, &ctl);
	if (rc >= 0)
		*val = ctl.value;

	return rc;
}
# endif

static int
sensor_set_picture(struct usb_ovfx2 *ov, struct video_picture *p)
{
//	int rc;

	PDEBUG(MISC|4, "sensor_set_picture");
# if 0
	rc = reg_w(ov, REG_CNTR, p->contrast >> 8);
	if (rc < 0)
		return rc;

	rc = reg_w(ov, REG_BRIGHT, (p->brightness >> 8) - 127);
	if (rc < 0)
		return rc;

	rc = reg_w(ov, REG_SAT, p->colour >> 8);
	if (rc < 0)
		return rc;

	rc = reg_w(ov, REG_HUE, (p->hue >> 8) - 127);
	if (rc < 0)
		return rc;
# endif
	return 0;
}

static int
sensor_get_picture(struct usb_ovfx2 *ov, struct video_picture *p)
{
//	int rc;

	PDEBUG(MISC|4, "sensor_get_picture");
# if 0
	/* Don't return error if a setting is unsupported, or rest of settings
         * will not be performed */

	rc = reg_r(ov, REG_CNTR);
	if (rc < 0)
		return rc;
	p->contrast = rc << 8;

	rc = reg_r(ov, REG_BRIGHT);
	if (rc < 0)
		return rc;
	p->brightness = (rc + 127) << 8;

	rc = reg_r(ov, REG_SAT);
	if (rc < 0)
		return rc;
	p->colour = rc << 8;

	rc = reg_r(ov, REG_HUE);
	if (rc < 0)
		return rc;
	p->hue = (rc + 127) << 8;

	p->whiteness = 105 << 8;
# endif
	return 0;
}

/* Matches the sensor's internal frame rate to the lighting frequency.
 * Valid frequencies are:
 *	50 - 50Hz, for European and Asian lighting
 *	60 - 60Hz, for American lighting
 *
 * Returns: 0 for success
 */
static int
sensor_set_light_freq(struct usb_ovfx2 *ov, int freq)
{
	if (freq != 50 && freq != 60) {
		err("Invalid light freq (%d Hz)", freq);
		return -EINVAL;
	}

	return sensor_set_control(ov, OVCAMCHIP_CID_FREQ, freq);
}

/* Returns number of bits per pixel (regardless of where they are located;
 * planar or not), or zero for unsupported format.
 */
static inline int
get_depth(int palette)
{
	switch (palette) {
	case VIDEO_PALETTE_RGB24:
		return 24; /* Planar */

	case VIDEO_PALETTE_RGB32:
		return 32; /* Planar */
		
	case VIDEO_PALETTE_RGB555:
	case VIDEO_PALETTE_RGB565:
		return 16; /* Packed planar */
		
	case VIDEO_PALETTE_YUV422:
	case VIDEO_PALETTE_YUYV:
		return 16; /* Packed Y U Y V Y U Y V (U and V have to be shared across two Y) */
		
	case VIDEO_PALETTE_UYVY:
		return 16; /* Packed U Y V Y U Y V Y (U and V have to be shared across two Y) */

	default:
		PDEBUG(MISC|4, "Palette unknown (%s)", symbolic(v4l1_plist, palette));
		return 0;  /* Invalid format */
	}
}

/* Bytes per frame. Used by read(). Return of 0 indicates error */
static inline long int
get_frame_length(struct ovfx2_frame *frame)
{
	if (!frame)
		return 0;
	else
		return ((frame->width * frame->height * frame->depth) >> 3);
}

static int
set_ov_sensor_window(struct usb_ovfx2 *ov, int width, int height, int mode,
		     int sub_flag)
{
	struct ovcamchip_window win;
	int rc;

	win.format = mode;

	/* Unless subcapture is enabled, center the image window and downsample
	 * if possible to increase the field of view */
	if (sub_flag) {
		win.x = ov->subx;
		win.y = ov->suby;
		win.width = ov->subw;
		win.height = ov->subh;
		win.quarter = 0;
	} else {
		win.x = 0;
		win.y = 0;
		win.width = width;
		win.height = height;
		win.quarter = 0;
	}

	if (clockdiv >= 0)
		win.clockdiv = clockdiv;
	else
		win.clockdiv = 0x01;	/* Always use the max frame rate */

	PDEBUG(MISC|4, "Calling sensor command");
	rc = sensor_cmd(ov, OVCAMCHIP_CMD_S_MODE, &win);
	if (rc < 0)
		return rc;
# ifdef HARDCODED_SENSOR
	if (framedrop >= 0)
		i2c_w(ov, 0x16, framedrop);
# endif
	return 0;
}

/* Sets up the OVFX2 with the given image parameters */
static int
mode_reinit_regs(struct usb_ovfx2 *ov, struct ovfx2_frame *frame, int sub_flag)
{
	int	height = frame->height;
	int	width = frame->width;
	
	if (sub_flag) {
		width = ov->subw;
		height = ov->subh;
	}

	PDEBUG(OUTGOING|3, "Set width:%d, height:%d, sub:%d",
	       width, height, sub_flag);

	if (width < ov->minwidth || height < ov->minheight) {
		err("Requested dimensions are too small");
		return -EINVAL;
	}

	/* Set palette */

	/* Set bus width */

	/* Set resolution */

	/* Set framerate/clock */
//	reg_w_mask(ov, 0x0f, 0x08, 0x08);

	/* Reinitialize (if necessary) */
	if (frame->reqdsensorstate != ov->sensorstate) {
		/* disable AGC & AEC */
		i2c_w_mask(ov, 0x13, 0x00, 0x05);
		
		if (frame->reqdsensorstate == SENSOR_UXGA) {
			i2c_w(ov, 0x12, 0x20);		/* UXGA mode (1600x1200) */
			reg_w_mask(ov, 0x0F, STOPSVGA, SVGAMASK);
			i2c_w(ov, 0x11, UXGA_SPEED);
  			
			/* reset windowing registers to get UXGA and SVGA to superimpose */
			i2c_w(ov, 0x18, 0xF3);
			i2c_w(ov, 0x17, 0x2B);
			i2c_w(ov, 0x32, 0x00);
			
			frame->bayer_decode = bayer_decode_BGBG_GRGR_8;
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB24;
			
			reg_w(ov, 0x02, FIFO8);		/* FIFO width select = 8 */
			i2c_w(ov, 0x00, UXGA_GAIN);
			PDEBUG(OUTGOING|FRAMESTATE|1, "Gone to UXGA, R2=0x%02X", reg_r(ov, 0x02));
		}
		else {
			i2c_w(ov, 0x12, 0x60);		/* SVGA mode (800x600) */
			reg_w_mask(ov, 0x0F, STARTSVGA, SVGAMASK);
			i2c_w(ov, 0x11, SVGA_SPEED);
			
			/* reset windowing registers to get UXGA and SVGA to superimpose */
			i2c_w(ov, 0x18, 0x78);
			i2c_w(ov, 0x17, 0x14);
			i2c_w(ov, 0x32, 0x05);
			
			frame->bayer_decode = bayer_decode_BGBG_GRGR_8; /* was 16 */
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB24;
			
			reg_w(ov, 0x02, FIFO8 /* was 16 */);	/* FIFO width select = 16 */
			i2c_w(ov, 0x00, SVGA_GAIN);
			PDEBUG(OUTGOING|FRAMESTATE|1, "Gone to SVGA, R2=0x%02X", reg_r(ov, 0x02));
		}
		
		ov->sensorstate = frame->reqdsensorstate;
	}		

	return 0;
}


/* This is a wrapper around the FX2 and sensor specific functions */
static int
mode_init_regs(struct usb_ovfx2 *ov,
	       int width, int height, int mode, int sub_flag)
{
	int rc = 0;
	
	reg_w(ov, 0x0f, 0x01);
	reg_w(ov, 0x0f, 0x03);
	reg_w(ov, 0x0f, 0x01);
	ov->last_slave = 0;		/* make sure we reset the i2c slave address before trying to write to sensor */
	reg_w(ov, 0x02, FIFO8);		/* set control-bit 0 (FIFO width=8) */
		
	PDEBUG(OUTGOING|1, "width:%d, height:%d, mode:%d, sub:%d",
	       width, height, mode, sub_flag);

	if (width < ov->minwidth || height < ov->minheight) {
		err("Requested dimensions are too small");
		return -EINVAL;
	}

	rc = set_ov_sensor_window(ov, width, height, mode, sub_flag);
	if (FATAL_ERROR(rc))
		return rc;
		
# if HARDCODED_SENSOR
	/********* The following settings are specific to this camera ********/
	i2c_w(ov, 0x03, 0x80);		  /* Hardcoded saturation value */
	i2c_w_mask(ov, 0x12, 0x48, 0x48); /* Enable h-mirror and raw output */
	i2c_w_mask(ov, 0x27, 0x10, 0x10); /* Bypass RGB matrix */
	i2c_w_mask(ov, 0x28, 0x04, 0x04); /* G/BR (YG) mode */
	i2c_w_mask(ov, 0x2a, 0x00, 0x10); /* No UV 2-pixel delay */
	i2c_w_mask(ov, 0x2d, 0x40, 0x40); /* QVGA 60 FPS (redundant?) */
	i2c_w(ov, 0x67, 0x00); 		  /* Disable YUV postprocess */
	i2c_w_mask(ov, 0x2d, 0x00, 0x0f); /* Disable anti-alias */
	i2c_w(ov, 0x24, 0x20); 		  /* Use QVGA white pixel ratio */
	i2c_w(ov, 0x25, 0x30); 		  /* Use QVGA black pixel ratio */
	i2c_w(ov, 0x60, 0x26);		  /* Unknown */
	i2c_w(ov, 0x74, 0x20);		  /* AGC max gain = 4x */
	i2c_w(ov, 0x06, 0x68);		  /* Hardcoded brightness value */
	i2c_w(ov, 0x14, 0x04);		  /*  */
# endif
	if (width > 800) {
		if (ov->sensorstate != SENSOR_UXGA) {
			i2c_w(ov, 0x12, 0x80 | 0x20); /* Reset to UXGA mode (1600x1200) */
			i2c_w(ov, 0x12, 0x20); /* UXGA mode (1600x1200) */
			reg_w_mask(ov, 0x0F, STOPSVGA, SVGAMASK);
			ov->frame[0].bayer_decode = bayer_decode_BGBG_GRGR_8;
			ov->frame[0].reqdsensorstate = SENSOR_UXGA;
			ov->frame[0].rematrix_even = ov->frame[0].rematrix_odd = Cam2RGB24;
			ov->frame[1].bayer_decode = bayer_decode_BGBG_GRGR_8;
			ov->frame[1].reqdsensorstate = SENSOR_UXGA;
			ov->frame[1].rematrix_even = ov->frame[1].rematrix_odd = Cam2RGB24;
			reg_w(ov, 0x02, FIFO8);	/* FIFO width select = 8 */
			ov->sensorstate = SENSOR_UXGA;
		}
	}
	else {
		if (ov->sensorstate != SENSOR_SVGA) {
			i2c_w(ov, 0x12, 0x80 | 0x60); /* Reset to SVGA mode (800x600) */
			i2c_w(ov, 0x12, 0x60); /* SVGA mode (800x600) */
			reg_w_mask(ov, 0x0F, STARTSVGA, SVGAMASK);
			ov->frame[0].bayer_decode = bayer_decode_BGBG_GRGR_8 /* was 16 */;
			ov->frame[0].reqdsensorstate = SENSOR_SVGA;
			ov->frame[0].rematrix_even = ov->frame[0].rematrix_odd = Cam2RGB24;
			ov->frame[1].bayer_decode = bayer_decode_BGBG_GRGR_8 /* was 16 */;
			ov->frame[1].reqdsensorstate = SENSOR_SVGA;
			ov->frame[1].rematrix_even = ov->frame[1].rematrix_odd = Cam2RGB24;
			reg_w(ov, 0x02, FIFO8 /* was 16 */); /* FIFO width select = 16 */
			ov->sensorstate = SENSOR_SVGA;
		}
	}

//	i2c_w(ov, 0x13, 0xCD); /* disable auto white balance */
//	i2c_w(ov, 0x11, 0x80); /* full speed, pll on */
			
	i2c_w(ov, 0x09, 0x01); /* normal reset time, double drive capacity */
	i2c_w(ov, 0x0D, 0x40); /* anti-blooming on */

	if (i2c_r(ov, 0x0B) == 0x41) {
		/* OV2610++ */
		i2c_w(ov, 0x33, 0xDC); /* set (double) ref current */
		i2c_w(ov, 0x35, 0x90); /* adjust anti-blooming */
		i2c_w(ov, 0x36, 0x37); /* adjust internal reference to "improve image" */
	}
	else {
		/* Ordinary OV2610 */
		i2c_w(ov, 0x33, 0xD0); /* set (double) ref current */
		i2c_w(ov, 0x35, 0x64); /* adjust anti-blooming */
		i2c_w(ov, 0x36, 0x17); /* adjust internal reference to "improve image" */
	}
	
//	i2c_w(ov, 0x11, 0x83); /* full speed, pll on, different divider (no idea why) */
	i2c_w(ov, 0x01, 0x8D); /* tweak blue gain for 5000K */
	i2c_w(ov, 0x02, 0x98); /* tweak red gain for 5000K */
	i2c_w(ov, 0x03, 0x40); /* low bits of blue and red gains for 5000K */
	i2c_w(ov, 0x2D, 0x00); /* VSYNC width (should be set to this already) */
	i2c_w(ov, 0x13, 0xC0); /* disable AGC, AEC & AWB */
	i2c_w(ov, 0x24, 0xB0); /* adjust AEC hi-water mark (default: 0xA0) */
	i2c_w(ov, 0x25, 0x90); /* adjust AEC low-water mark (default: 0x88) */
	i2c_w(ov, 0x10, 0x43); /* exposure value (should be set to this already) */
	
	/* Our own tweaks */
//	i2c_w_mask(ov, 0x12, 0x02, 0x02); /* Colour Bar enable */

	if (width > 800) {
		i2c_w(ov, 0x11, UXGA_SPEED);
		i2c_w(ov, 0x00, UXGA_GAIN);
	}
	else {
		i2c_w(ov, 0x11, SVGA_SPEED);
			
		/* reset windowing registers to get UXGA and SVGA to superimpose */
		i2c_w(ov, 0x18, 0x78);
		i2c_w(ov, 0x17, 0x14);
		i2c_w(ov, 0x32, 0x05);
		
		i2c_w(ov, 0x00, SVGA_GAIN);
	}
	
	return 0;
}

/* Set up the camera chip with the options provided by the module params */
static int
camchip_init_settings(struct usb_ovfx2 *ov)
{
	int rc;

	rc = sensor_set_control(ov, OVCAMCHIP_CID_AUTOBRIGHT, autobright);
	if (FATAL_ERROR(rc))
		return rc;

	rc = sensor_set_control(ov, OVCAMCHIP_CID_AUTOEXP, autoexp);
	if (FATAL_ERROR(rc))
		return rc;

	rc = sensor_set_control(ov, OVCAMCHIP_CID_BANDFILT, bandingfilter);
	if (FATAL_ERROR(rc))
		return rc;

	if (lightfreq) {
		rc = sensor_set_light_freq(ov, lightfreq);
		if (FATAL_ERROR(rc))
			return rc;
	}

	rc = sensor_set_control(ov, OVCAMCHIP_CID_BACKLIGHT, backlight);
	if (FATAL_ERROR(rc))
		return rc;

	rc = sensor_set_control(ov, OVCAMCHIP_CID_MIRROR, mirror);
	if (FATAL_ERROR(rc))
		return rc;

	return 0;
}


/**********************************************************************
 *
 * Frame buffering
 *
 **********************************************************************/
# if 0
/* Sleeps until no frames are active. Returns !0 if got signal or unplugged */
static int
ovfx2_wait_frames_inactive(struct usb_ovfx2 *ov)
{
	int rc=0;

/*	rc = wait_event_interruptible(ov->wq,
			ov->curframe == -1 ||
			ov->frame[ov->curframe].state == FREE_EMPTY ||
			ov->frame[ov->curframe].state == FREE_FULL ||
			!ov->present);
 */		
	PDEBUG(FRAMESTATE|3, "ov->curframe=%d ov->frame[ov->curframe].state=%d ov->present=%d",
			ov->curframe, ov->frame[ov->curframe].state, ov->present);
	
	if (rc)
		return rc;

	if (ov->present) {
		ovfx2_pause(ov);
		
		for (rc = 0; rc < OVFX2_NUMFRAMES; rc++) {
			/* abandon all filling or filled frames */
			switch (ov->frame[rc].state) {
			case FREE_FILLING:
			case FREE_FULL:
			case FRAME_ERROR:
				ov->frame[rc].state = FREE_EMPTY;
				break;
			
			case MARKED_FILLING:
			case MARKED_FULL:
				ov->frame[rc].state = MARKED_EMPTY;
				break;
				
			default:;
			}
			
			/* notify anyone who might have been waiting */
			wake_up_interruptible(&ov->frame[rc].wq);
		}
		
		return 0;
	}

	return -ENODEV;
}
# endif

/**********************************************************************
 *
 * Raw data parsing
 * The Bayer-processing code invokes the following routines to pack raw camera R'G'B' signals into various FourCC formats.
 *
 **********************************************************************/

/*
 * The following three routines correct for the camera primaries AND derive RGB from them.
 * These routines have to be inlined - you can't take the hit of a real subroutine call.
 *
 * Caveat: The equations are based on Omnivision's documented matrix for deriving RGB from the
 * camera primaries. I don't know the output RGB colorimetry of Omnivision's matrix - SH.
 *
 * Double caveat: Is the RGB colorimetry of V4L specified anywhere?
 */	
static inline unsigned char
Matrix_to_R( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int R;
	
	/* R = 1.315R' - 0.165G' - 0.150B' */
	if ((R = ((rdash*86180 - gdash*10813 - bdash*9830 + 32768) >> 16)) > 0xFF)
		R = 0xFF;
	else if (R < 0)
		R = 0;
	
	return R;
}


static inline unsigned char
Matrix_to_G( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int G;
	
	/* G = -0.09R' + 1.205G' - 0.115B' */
	if ((G = ((-rdash*5898 + gdash*78971 - bdash*7537 + 32768) >> 16)) > 0xFF)
		G = 0xFF;
	else if (G < 0)
		G = 0;
	
	return G;
}


static inline unsigned char
Matrix_to_B( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int B;
	
	/* B = 0.021R' - 1.013G' + 1.992B' */
	if ((B = ((rdash*1376 - gdash*66388 + bdash*130548 + 32768) >> 16)) > 0xFF)
		B = 0xFF;
	else if (B < 0)
		B = 0;
	
	return B;
}


static unsigned char *
Cam2RGB24( unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{

	*pOut++ = Matrix_to_B(rdash, gdash, bdash);		/* LSB */
	*pOut++ = Matrix_to_G(rdash, gdash, bdash);
	*pOut++ = Matrix_to_R(rdash, gdash, bdash);		/* MSB */
	return pOut;
}


static unsigned char *
Cam2RGB32( unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{

	*pOut++ = Matrix_to_B(rdash, gdash, bdash);		/* LSB */
	*pOut++ = Matrix_to_G(rdash, gdash, bdash);
	*pOut++ = Matrix_to_R(rdash, gdash, bdash);
	*pOut++ = 0;						/* MSB */
	return pOut;
}


static unsigned char *
Cam2RGB565( unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	unsigned char r, g, b;
	
	b = Matrix_to_B(rdash, gdash, bdash);
	g = Matrix_to_G(rdash, gdash, bdash);
	r = Matrix_to_R(rdash, gdash, bdash);
		
	/*             +---+---+---+---+---+---+---+---+       +---+---+---+---+---+---+---+---+
	 *        MSB: | R | R | R | R | R | G | G | G |  LSB: | G | G | G | B | B | B | B | B |
	 *             +---+---+---+---+---+---+---+---+       +---+---+---+---+---+---+---+---+
	 */
	*pOut++ = ((g << 3) & 0xE0) | ((b >> 3) & 0x1F);		/* LSB */
	*pOut++ = (r & 0xF8) | (g >> 5);				/* MSB */
	return pOut;
}


static unsigned char *
Cam2RGB555( unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	unsigned char r, g, b;
	
	b = Matrix_to_B(rdash, gdash, bdash);
	g = Matrix_to_G(rdash, gdash, bdash);
	r = Matrix_to_R(rdash, gdash, bdash);
		
	/*             +---+---+---+---+---+---+---+---+       +---+---+---+---+---+---+---+---+
	 * First byte: | 0 | R | R | R | R | R | G | G |  LSB: | G | G | G | R | R | R | R | R |
	 *             +---+---+---+---+---+---+---+---+       +---+---+---+---+---+---+---+---+
	 */
	*pOut++ = ((g << 2) & 0xE0) | ((b >> 3) & 0x1F);		/* LSB */
	*pOut++ = ((r >> 1) & 0x7C) | (g >> 6);				/* MSB */
	return pOut;
}


/*
 * The following three routines correct for the camera primaries AND derive Rec.601 YUV from them.
 * These routines have to be inlined - you can't take the hit of a real subroutine call.
 *
 * Caveat: The equations are based on Omnivision's documented matrix for deriving RGB from the
 * camera primaries. I don't know for sure that the output colorimetry of Omnivision's matrix
 * is actually Rec.601 compliant - SH.
 *
 * Double caveat: Is the RGB colorimetry of V4L specified anywhere? Or the assumed YUV matrixing?
 */	
static inline unsigned char
Matrix_to_Y( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int Y;
	
	/* ideally: Y = 0.3427R' +  0.5425G' +  0.1147B' */
	if ((Y = ((22459*rdash + 35553*gdash + 7517*bdash + 32768) >> 16)) > 0xFF)
		Y = 0xFF;
	else if (Y < 0)
		Y = 0;
	
	return Y;
}


static inline unsigned char
Matrix_to_U( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int U;
	
	/* ideally: U = -0.1816R' - 0.8778G' + 1.0594B' */
	if ((U = ((69429*bdash - 57528*gdash - 11901*rdash + 32768) >> 16) + 0x80) > 0xFF)
		U = 0xFF;
	else if (U < 0)
		U = 0;
	
	return U;
}


static inline unsigned char
Matrix_to_V( unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
	int V;
	
	/* ideally: V = 0.6935R' - 0.5046G' - 0.1888B' */
	if ((V = ((45449*rdash - 33069*gdash - 12373*bdash + 32768) >> 16) + 0x80) > 0xFF)
		V = 0xFF;
	else if (V < 0)
		V = 0;
	
	return V;
}


static unsigned char *
Cam2YUYV_Even(unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
		
	*pOut++ = Matrix_to_Y(rdash, gdash, bdash);		/* 1st byte */
	*pOut++ = Matrix_to_U(rdash, gdash, bdash);		/* 2nd byte */
	return pOut;
}


static unsigned char *
Cam2YUYV_Odd(unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
		
	*pOut++ = Matrix_to_Y(rdash, gdash, bdash);		/* 1st byte */
	*pOut++ = Matrix_to_V(rdash, gdash, bdash);		/* 2nd byte */
	return pOut;
}


static unsigned char *
Cam2UYVY_Even(unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
		
	*pOut++ = Matrix_to_U(rdash, gdash, bdash);		/* 1st byte */
	*pOut++ = Matrix_to_Y(rdash, gdash, bdash);		/* 2nd byte */
	return pOut;
}


static unsigned char *
Cam2UYVY_Odd(unsigned char *pOut, unsigned char rdash, unsigned char gdash, unsigned char bdash )
{
		
	*pOut++ = Matrix_to_V(rdash, gdash, bdash);		/* 1st byte */
	*pOut++ = Matrix_to_Y(rdash, gdash, bdash);		/* 2nd byte */
	return pOut;
}


# if 0
/* Flip image vertically */
static void
vert_mirror(unsigned char *pIn,
	    unsigned char *pOut,
	    unsigned int width,
	    unsigned int height,
	    unsigned int depth)
{
	const unsigned int linesize = width * depth;
	int i;

	pOut += height * linesize;

	for (i = 0; i < height; i++) {
		pOut -= linesize;
		memcpy(pOut, pIn, linesize);
		pIn += linesize;
	}
}
# endif

/**********************************************************************
 *
 * Image processing
 *
 **********************************************************************/

static void
ovfx2_postprocess(struct usb_ovfx2 *ov, struct ovfx2_frame *frame)
{

	if (dumppix == 1) {		/* Dump with color decoding */
		memset(frame->data, 0,
			MAX_DATA_SIZE(ov->maxwidth, ov->maxheight));
		PDEBUG(FRAMESTATE|INCOMING|4, "Dumping %d bytes", frame->bytes_recvd);
		frame->bayer_decode(frame->data, frame->rawdata, frame->rawwidth, frame->rawheight, 0, frame->rematrix_even, frame->rematrix_odd);
	} else if (dumppix == 2) {	/* Dump raw data */
		memset(frame->data, 0,
			MAX_DATA_SIZE(ov->maxwidth, ov->maxheight));
		PDEBUG(FRAMESTATE|INCOMING|4, "Dumping %d bytes", frame->bytes_recvd);
		memcpy(frame->data, frame->rawdata, frame->bytes_recvd);
	} else {
		/* FIXME: Remove this once decoding is stable */
/*		PDEBUG(FRAMESTATE|INCOMING|4, "Filling %d bytes", MAX_DATA_SIZE(frame->width, frame->height)); */
		frame->bayer_decode(frame->data, frame->rawdata, frame->rawwidth, frame->rawheight, 0, frame->rematrix_even, frame->rematrix_odd);
	}
}

/**********************************************************************
 *
 * OV2610 data transfer, IRQ handler
 *
 **********************************************************************/

static inline void
ovfx2_move_data(struct usb_ovfx2 *ov, unsigned char *in, int n)
{
	int max_raw = MAX_RAW_DATA_SIZE(ov->maxwidth, ov->maxheight);
	struct ovfx2_frame *frame = &ov->frame[ov->curframe];
	struct timeval *ts;
    	int nextf;

	switch (frame->state) {
	case FREE_EMPTY:
		if (ov->fsync == 1) {
			PDEBUG(FRAMESTATE|1, "Starting capture on free frame %d", frame->framenum);
			frame->state = FREE_FILLING;
			frame->bytes_recvd = 0;
			frame->reqdsensorstate = ov->sensorstate;
			ov->fsync = 0;
		}
		else {
			PDEBUG(FRAMESTATE|4, "Waiting for frame-sync on free frame %d", frame->framenum);
			if (n < OVFX2_BULK_SIZE)
				ov->fsync++;
				
			return;
		}
		
		break;

	case MARKED_EMPTY:
		if (frame->reqdsensorstate != ov->sensorstate) {
			/* tricky: must change sensor state before filling */
			PDEBUG(FRAMESTATE|1, "Must change mode on marked frame %d", frame->framenum);
			ov->fx2state = FX2_WRONGMODE;
			wake_up_interruptible(&frame->wq);
			wake_up_interruptible(&ov->wq);
			ov->fsync = -2;		/* lose two frames after the switch */
			return;
		}
		else if (ov->fsync == 1) {
			PDEBUG(FRAMESTATE|1, "Starting capture on marked empty frame %d", frame->framenum);
			frame->state = MARKED_FILLING;
			frame->bytes_recvd = 0;
			ov->fsync = 0;
		}
		else {
			PDEBUG(FRAMESTATE|4, "Waiting for frame-sync on marked empty frame %d", frame->framenum);
			if (n < OVFX2_BULK_SIZE)
				ov->fsync++;
				
			return;
		}
		
		break;
	
	case MARKED_FULL:
	case FREE_FULL:
		if (ov->fsync == 1) {
			/* We need to start a new frame. If next frame is usable, point to it */
			nextf = (ov->curframe + 1) % OVFX2_NUMFRAMES;
			ov->fsync = 0;
		
			switch (ov->frame[nextf].state) {
			case FREE_EMPTY:
				PDEBUG(FRAMESTATE|3, "Pointing to vacant frame %d", nextf);
				ov->frame[ov->curframe].reqdsensorstate = ov->sensorstate;
				goto take_nextf;
			
			case MARKED_EMPTY:
				PDEBUG(FRAMESTATE|3, "Pointing to empty (marked) frame %d", nextf);
				ov->curframe = nextf;
				frame = &ov->frame[ov->curframe];
				frame->state = MARKED_FILLING;
				frame->bytes_recvd = 0;
				break;
			
			case FREE_FULL:
				/* looks like we've run right round the block, but we can re-use this one */
				PDEBUG(FRAMESTATE|3, "No true empty frames left - refilling %d", nextf);
		
			take_nextf:
				/* when we fill a free, unmarked frame we guess that when it does get marked, it will not require a mode-switch */
				ov->curframe = nextf;
				frame = &ov->frame[ov->curframe];
				frame->reqdsensorstate = ov->sensorstate;
				frame->state = FREE_FILLING;
				frame->bytes_recvd = 0;
				break;
			
			case MARKED_FULL:
				/* looks like we've run right round the block */
				if (frame->state == FREE_FULL) {
					/* we'll have to re-use the one we just filled */
					PDEBUG(FRAMESTATE|3, "Can't move on - refilling %d", ov->curframe);
					frame->state = FREE_FILLING;
					frame->bytes_recvd = 0;
				}
				else {
					PDEBUG(FRAMESTATE|3, "No empty frames left: pausing");
					ov->curframe = -1;
					return;
				}
			
				break;
			
			default:
				/* this "can't happen" :-) */
				PDEBUG(FRAMESTATE|3, "Frame %d not ready. State = %d", nextf, ov->frame[nextf].state);
			
				/* stop grabbing */
				ov->curframe = -1;
				return;
			}
		}
		else {
			/* we'll delay until we really are at the start of a frame */
			PDEBUG(FRAMESTATE|3, "Waiting for frame-sync before moving on from %d", frame->framenum);
			if (n < OVFX2_BULK_SIZE)
				ov->fsync++;
				
			return;
		}
		
		break;
		
	default:
		/* the other states don't involve starting a frame */
		break;
	}
	
	if (n) {
		int b = n;	/* Number of image data bytes */
# ifdef HARDCODED_SENSOR
		/* The first 2560 bytes of data (first 2 lines?) are invalid */
		if (frame->bytes_recvd == 0) {
			in += 2560;
			b -= 2560;
		}
#endif
		frame->bytes_recvd += b;
		if (frame->bytes_recvd <= max_raw)
			memcpy(frame->rawdata + frame->bytes_recvd - b, in, b);
		else
			PDEBUG(INCOMING|5, "Raw data buffer overrun!! (%d)",
				frame->bytes_recvd - max_raw);
	} else {
		PDEBUG(INCOMING|4, "Empty bulk packet");
	}

	/* Short packets indicate end-of-frame */
	if (n < OVFX2_BULK_SIZE) {
		PDEBUG(INCOMING|3, "Frame end, curframe = %d, recvd=%d",
			ov->curframe, frame->bytes_recvd);

		/* Don't allow byte count to exceed buffer size */
		RESTRICT_TO_RANGE(frame->bytes_recvd, 8, max_raw);

		ts = (struct timeval *)(frame->data
		      + MAX_FRAME_SIZE(ov->maxwidth, ov->maxheight));
		do_gettimeofday(ts);

		// FIXME: Since we don't know the header formats yet,
		// there is no way to know what the actual image size is
		frame->rawwidth = frame->width;
		frame->rawheight = frame->height;

		/* Don't wake up user for same frame more than once */
		if (frame->state == MARKED_FILLING) {
			if (frame->reqdsensorstate == SENSOR_SVGA && frame->bytes_recvd != 800*600 /* extra "*2" in FIFO16 mode */) {
				PDEBUG(FRAMESTATE|2, "Frame %d wrong size (%d)", ov->curframe, frame->bytes_recvd);
				ov->fx2state = FX2_WRONGSIZE;
				wake_up_interruptible(&frame->wq);
				wake_up_interruptible(&ov->wq);
				frame->bytes_recvd = 0;
			}
			else if (frame->reqdsensorstate == SENSOR_UXGA && frame->bytes_recvd != 1600*1200) {
				PDEBUG(FRAMESTATE|2, "Frame %d wrong size (%d)", ov->curframe, frame->bytes_recvd);
				ov->fx2state = FX2_WRONGSIZE;
				wake_up_interruptible(&frame->wq);
				wake_up_interruptible(&ov->wq);
				frame->bytes_recvd = 0;
			}
			else {
				PDEBUG(FRAMESTATE|2, "Tell user about frame %d", ov->curframe);
				frame->state = MARKED_FULL;
				wake_up_interruptible(&frame->wq);
			}
		}
		else if (frame->state == FREE_FILLING)
			frame->state = FREE_FULL;
		
		ov->fsync++;
	}
	
	return;
}


/* for bomb diagnostics: */
static int last_actlen = 0;

static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 51)
ovfx2_bulk_irq(struct urb *urb, struct pt_regs *regs)
#else
ovfx2_bulk_irq(struct urb *urb)
#endif
{
	int i;
	int actlen = urb->actual_length;
	struct usb_ovfx2 *ov;
	struct ovfx2_sbuf *sbuf;

	if (!urb->context) {
		PDEBUG(INCOMING|4, "no context");
		return;
	}

	sbuf = urb->context;
	ov = sbuf->ov;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov || !interface_to_usbdev(ov->intf) || ov->rawfbuf == NULL) {
#else
	if (!ov || !ov->dev || ov->rawfbuf == NULL) {
#endif
		PDEBUG(INCOMING|4, "no device, or not open");
		return;
	}

        if (urb->status == -ENOENT || urb->status == -ECONNRESET) {
                PDEBUG(INCOMING|4, "URB unlinked");
                return;
        }
	
	if ( !(ov->fx2state == FX2_WRONGMODE || ov->fx2state == FX2_STREAMING)) {
		PDEBUG(INCOMING|4, "hmmm... not streaming, but got interrupt");
		return;
	}

	if (urb->status != -EINPROGRESS && urb->status != 0) {
		err("ERROR: urb->status=%d: %s", urb->status,
		    symbolic(urb_errlist, urb->status));
	}
	else if (ov->bomb_running) {
		/* normal packet received: reset bomb */
		mod_timer(&ov->timebomb, jiffies+(HZ/10));

		PDEBUG(INCOMING|5, "sbuf[%d]: Got bulk completion: actlen=%d", sbuf->n, actlen);

		/* Warning: Don't copy data if no frame active! */
		if (ov->curframe >= 0 && ov->fx2state == FX2_STREAMING) {
			ovfx2_move_data(ov, urb->transfer_buffer, actlen);
			last_actlen = actlen;
		}
		else if (waitqueue_active(&ov->wq)) {
			wake_up_interruptible(&ov->wq);
		}
	}

	/* Resubmit this URB */
	urb->status = 0; // FIXME: Is this necessary? What about w/ OV511/OV518?
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	urb->dev = interface_to_usbdev(ov->intf);
#else
	urb->dev = ov->dev;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 4)
	if ((i = usb_submit_urb(urb, GFP_ATOMIC)) != 0)
#else
	if ((i = usb_submit_urb(urb)) != 0)
#endif
		err("usb_submit_urb() ret %d", i);

	return;
}

static void	timebomb_irq( unsigned long val );

	
/****************************************************************************
 *
 * Stream initialization and termination
 *
 ***************************************************************************/

static int
ovfx2_start_stream(struct usb_ovfx2 *ov)
{
	struct urb *urb;
	int err, n, size;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	struct usb_device *dev = interface_to_usbdev(ov->intf);
#else
	struct usb_device *dev = ov->dev;
#endif

	PDEBUG(FNCS|2, "Entering ovfx2_start_stream()");
	PDEBUG(OUTGOING|3, "*** Starting video stream ***");

	size = OVFX2_BULK_SIZE;

	for (n = 0; n < OVFX2_NUMSBUF; n++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 5)
		urb = usb_alloc_urb(0, GFP_KERNEL);
#else
		urb = usb_alloc_urb(0);
#endif
		if (!urb) {
			err("start stream: usb_alloc_urb ret. NULL");
			PDEBUG(FNCS|2, "Exiting ovfx2_start_stream()");
			return -ENOMEM;
		}

		usb_fill_bulk_urb(urb,
				  dev,
				  usb_rcvbulkpipe(dev,
						  OVFX2_BULK_ENDPOINT_ADDR),
				  ov->sbuf[n].data,
				  size,
				  ovfx2_bulk_irq,
				  &ov->sbuf[n]);

		ov->sbuf[n].urb = urb;
	}

	for (n = 0; n < OVFX2_NUMSBUF; n++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 4)
		err = usb_submit_urb(ov->sbuf[n].urb, GFP_KERNEL);
#else
		err = usb_submit_urb(ov->sbuf[n].urb);
#endif
		if (err) {
			err("start stream: usb_submit_urb(%d) ret %d", n, err);
			PDEBUG(FNCS|2, "Exiting ovfx2_start_stream()");
			return err;
		}
	}

	/* Start stream */
	err = reg_w_mask(ov, 0x0f, STARTSTREAM, STREAMMASK);
	if (err) {
		PDEBUG(FNCS|2, "Exiting ovfx2_start_stream()");
		return err;
	}

	ov->fx2state = FX2_STREAMING;
	ov->fsync = -1;		/* lose one frame on start */

	if (ov->bomb_running == 0) {
		init_timer(&ov->timebomb);
		ov->timebomb.expires = jiffies + (HZ/10);	/* tenth second */
		ov->timebomb.data = ( unsigned long ) ov;
		ov->timebomb.function = timebomb_irq;
		add_timer(&ov->timebomb);
		PDEBUG(BOMB|3, "*** Bomb set ***");
		ov->bomb_running = 1;
	}
	
	PDEBUG(FNCS|2, "Exiting ovfx2_start_stream()");
	return 0;
}

static void
ovfx2_unlink_bulk(struct usb_ovfx2 *ov)
{
	int n;

	PDEBUG(FNCS|2, "Entering ovfx2_unlink_bulk()");

	/* Unschedule all of the bulk td's */
	for (n = OVFX2_NUMSBUF - 1; n >= 0; n--) {
		if (ov->sbuf[n].urb) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 10)
			usb_kill_urb(ov->sbuf[n].urb);
#else
			usb_unlink_urb(ov->sbuf[n].urb);
#endif
			usb_free_urb(ov->sbuf[n].urb);
			ov->sbuf[n].urb = NULL;
		}
	}
	
	PDEBUG(FNCS|2, "Exiting ovfx2_unlink_bulk()");
	return;
}

static int
ovfx2_stop_stream(struct usb_ovfx2 *ov)
{
	int ret = 0;
	
//	if (ov->fx2state != FX2_STREAMING) {
//		PDEBUG(FNCS|2, "Repelled from ovfx2_stop_stream()");
//		return -ENODEV;
//	}

	PDEBUG(FNCS|2, "Entering ovfx2_stop_stream()");
	PDEBUG(OUTGOING|3, "*** Stopping capture ***");

	if (ov->bomb_running) {
		del_timer(&ov->timebomb);
		ov->bomb_running = 0;
	}
	
	ov->fx2state = FX2_NOT_STREAMING;

	/* Stop stream on FX2 */
	reg_w_mask(ov, 0x0f, STOPSTREAM, STREAMMASK);
	ret = reg_w(ov, 0x02, FIFO8);		/* FIFO width select = 8 */
	ovfx2_unlink_bulk(ov);

	PDEBUG(FNCS|2, "Exiting ovfx2_stop_stream()");
	return ret;
}

static int
ovfx2_new_frame(struct usb_ovfx2 *ov, int framenum, int format )
{
	struct ovfx2_frame *frame = &ov->frame[framenum];
	enum sensor_state	reqdsensorstate;

	PDEBUG(FRAMESTATE|1, "%s request: framenum = %d, width = %d, height = %d", (format == -1)? "Repeat": "New", framenum,
	       frame->width, frame->height);
	PDEBUG(FRAMESTATE|3, "Current state of framenum %d is %d", framenum, frame->state);
			
	if (frame->width > 800 || frame->height > 600) {
		reqdsensorstate = SENSOR_UXGA;
		frame->bayer_decode = bayer_decode_BGBG_GRGR_8;
	}
	else {
		reqdsensorstate = SENSOR_SVGA;
		frame->bayer_decode = bayer_decode_BGBG_GRGR_8 /* 16 */;
	}
	
	if (format != -1) {
		if (format == VIDEO_PALETTE_YUV422 ||
		    format == VIDEO_PALETTE_YUYV) {
			frame->rematrix_even = Cam2YUYV_Even;
			frame->rematrix_odd = Cam2YUYV_Odd;
		}
		if (format == VIDEO_PALETTE_UYVY) {
			frame->rematrix_even = Cam2UYVY_Even;
			frame->rematrix_odd = Cam2UYVY_Odd;
		}
		else if (format == VIDEO_PALETTE_RGB24)
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB24;
		else if (format == VIDEO_PALETTE_RGB32)
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB32;
		else if (format == VIDEO_PALETTE_RGB565)
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB565;
		else if (format == VIDEO_PALETTE_RGB555)
			frame->rematrix_even = frame->rematrix_odd = Cam2RGB555;
 	}

	switch (frame->state) {
	case FREE_EMPTY:
	case FRAME_ERROR:
//	case FRAME_ABORT:
		PDEBUG(FRAMESTATE|3, "promote to MARKED_EMPTY");
		frame->state = MARKED_EMPTY;
		break;
	
	case FREE_FILLING:
		if (frame->reqdsensorstate == reqdsensorstate) {
			PDEBUG(FRAMESTATE|3, "promote to MARKED_FILLING");
			frame->state = MARKED_FILLING;
		}
		else {
			/* it's filling, but in the wrong mode */
			PDEBUG(FRAMESTATE|3, "demote to MARKED_EMPTY");
			frame->state = MARKED_EMPTY;
		}
		
		break;
	
	case FREE_FULL:
		if (frame->reqdsensorstate == reqdsensorstate) {
			PDEBUG(FRAMESTATE|3, "promote to MARKED_FULL");
			frame->state = MARKED_FULL;
		}
		else {
			/* it's full, but in the wrong mode */
			PDEBUG(FRAMESTATE|3, "demote to MARKED_EMPTY");
			frame->state = MARKED_EMPTY;
		}
		break;
	
	default:
		PDEBUG(FRAMESTATE|3, "nothing to do: state=%d", frame->state);
		return 0;
	}
	
	/* the sensor's state can't be toggled until this frame is ready to be filled */
	/* (and that happens in the bulk transfer callback routine) */
	frame->reqdsensorstate = reqdsensorstate;

	if (ov->curframe == -1)
		ov->curframe = framenum;

	/* Make sure it's not too big */
	if (frame->width > ov->maxwidth)
		frame->width = ov->maxwidth;

	frame->width &= ~7L;		/* Multiple of 8 */

	if (frame->height > ov->maxheight)
		frame->height = ov->maxheight;

	frame->height &= ~3L;		/* Multiple of 4 */

	return 0;
}


static void
timebomb_irq( unsigned long val )
{
	struct usb_ovfx2 *ov = ( struct usb_ovfx2 * ) val;
	int		n;
	
	PDEBUG(BOMB|2, "*** Bomb went off! ***");
	ov->fx2state = FX2_WEDGED;
	
	if (ov->curframe != -1)
		/* abandon the frame we were filling */
		ov->frame[ov->curframe].state = FRAME_ERROR;

	/* This may cause the user process to request another frame */
	for (n = 0; n < OVFX2_NUMFRAMES; n++)
		wake_up_interruptible(&ov->frame[n].wq);

	wake_up_interruptible(&ov->wq);
	return;
}

/****************************************************************************
 *
 * Buffer management
 *
 ***************************************************************************/

/*
 * - This code will dealloc the frame-buffer which may have been MMAP-ed.
 * - You must only call this if you're processing the final close() on the
 *   file-descriptor!
 */
static void
ovfx2_dealloc_fbuf(struct usb_ovfx2 *ov)
{

	PDEBUG(FNCS|3, "Entering ovfx2_dealloc_fbuf()");
	down(&ov->buf_lock);

	if (ov->fbuf) {
		rvfree(ov->fbuf, OVFX2_NUMFRAMES * MAX_DATA_SIZE(ov->maxwidth, ov->maxheight));
		ov->fbuf = NULL;
	}
	
	up(&ov->buf_lock);
	PDEBUG(FNCS|3, "Exiting ovfx2_dealloc_fbuf()");
	return;
}


/*
 * - You must acquire buf_lock before entering this function.
 * - Because this code will free any non-null pointer, you must be sure to null
 *   them if you explicitly free them somewhere else! This routine doesn't dealloc
 *   'fbuf' as it might be MMAP-ed.
 */
static void
ovfx2_do_dealloc(struct usb_ovfx2 *ov)
{
	int i;

	PDEBUG(FNCS|3, "Entering ovfx2_do_dealloc()");

	if (ov->rawfbuf) {
		vfree(ov->rawfbuf);
		ov->rawfbuf = NULL;

		/*
		 * Nulling ov->rawfbuf may cause the user process to abandon any blocked
		 * CSYNC ioctl or read() call.
		 */
		for (i = 0; i < OVFX2_NUMFRAMES; i++)
			wake_up_interruptible(&ov->frame[i].wq);

		wake_up_interruptible(&ov->wq);
	}

# ifdef TBUF
	if (ov->tempfbuf) {
		vfree(ov->tempfbuf);
		ov->tempfbuf = NULL;
	}
# endif

	for (i = 0; i < OVFX2_NUMSBUF; i++) {
		if (ov->sbuf[i].data) {
			kfree(ov->sbuf[i].data);
			ov->sbuf[i].data = NULL;
			PDEBUG(MISC|3, "ov->sbuf[%d].data free", i);
		}
	}

	for (i = 0; i < OVFX2_NUMFRAMES; i++) {
		ov->frame[i].data = NULL;
		ov->frame[i].rawdata = NULL;
# ifdef TBUF
		ov->frame[i].tempdata = NULL;
# endif
	}

	PDEBUG(MISC|4, "buffer memory deallocated");
	ov->buf_state = BUF_NOT_ALLOCATED;
	PDEBUG(FNCS|3, "Exiting ovfx2_do_dealloc()");
	return;
}

static int
ovfx2_alloc(struct usb_ovfx2 *ov)
{
	int i;
	const int w = ov->maxwidth;
	const int h = ov->maxheight;
	const int data_bufsize = OVFX2_NUMFRAMES * MAX_DATA_SIZE(w, h);
	const int raw_bufsize = OVFX2_NUMFRAMES * MAX_RAW_DATA_SIZE(w, h);

	PDEBUG(FNCS|2, "Entering ovfx2_alloc()");
	down(&ov->buf_lock);

	if (ov->buf_state == BUF_ALLOCATED)
		goto out;

	ov->fbuf = rvmalloc(data_bufsize);
	if (!ov->fbuf)
		goto error;

	ov->rawfbuf = vmalloc(raw_bufsize);
	if (!ov->rawfbuf)
		goto error;

	memset(ov->rawfbuf, 0, raw_bufsize);

# ifdef TBUF
	ov->tempfbuf = vmalloc(raw_bufsize);
	if (!ov->tempfbuf)
		goto error;

	memset(ov->tempfbuf, 0, raw_bufsize);
# endif
	for (i = 0; i < OVFX2_NUMSBUF; i++) {
		ov->sbuf[i].data = kmalloc(OVFX2_BULK_SIZE, GFP_KERNEL);
		if (!ov->sbuf[i].data)
			goto error;

		PDEBUG(MISC|3, "ov->sbuf[%d] @ %p", i, ov->sbuf[i].data);
	}

	for (i = 0; i < OVFX2_NUMFRAMES; i++) {
		ov->frame[i].data = ov->fbuf + i * MAX_DATA_SIZE(w, h);
		ov->frame[i].rawdata = ov->rawfbuf
		 + i * MAX_RAW_DATA_SIZE(w, h);
# ifdef TBUF
		ov->frame[i].tempdata = ov->tempfbuf
		 + i * MAX_RAW_DATA_SIZE(w, h);
# endif
		PDEBUG(MISC|3, "ov->frame[%d] @ %p", i, ov->frame[i].data);
	}

	ov->buf_state = BUF_ALLOCATED;
	
out:
	up(&ov->buf_lock);
	PDEBUG(FNCS|2, "Exiting ovfx2_alloc(): OK");
	return 0;
	
error:
	ovfx2_do_dealloc(ov);
	up(&ov->buf_lock);
	PDEBUG(FNCS|2, "Exiting ovfx2_alloc(): error");
	return -ENOMEM;
}

static void
ovfx2_dealloc(struct usb_ovfx2 *ov)
{

	PDEBUG(FNCS|2, "Entering ovfx2_dealloc()");
	down(&ov->buf_lock);
	ovfx2_do_dealloc(ov);
	up(&ov->buf_lock);
	PDEBUG(FNCS|2, "Exiting ovfx2_dealloc()");
	return;
}

	
static void
ovfx2_cbuf_dealloc(struct usb_ovfx2 *ov)
{

	/* let go the command-buffer */
	PDEBUG(FNCS|2, "Entering ovfx2_cbuf_dealloc()");
	down(&ov->cbuf_lock);
	
	if (ov->cbuf) {
		kfree(ov->cbuf);
		ov->cbuf = NULL;
		PDEBUG(MISC|2, "ov->cbuf free");
	}
	
	up(&ov->cbuf_lock);
	PDEBUG(FNCS|2, "Exiting ovfx2_cbuf_dealloc()");
	return;
}

/****************************************************************************
 *
 * Basic V4L API functionality
 *
 ***************************************************************************/

int
ovfx2_initial_open( struct usb_ovfx2 *ov, struct file *file )
{
	int err, i;

	PDEBUG(FNCS|1, "Entering ovfx2_initial_open()");
	
	/* only do this stuff on the first open */
	down(&ov->lock);
	
	ov->sub_flag = 0;
	ov->currentformat = (force_palette)? force_palette: VIDEO_PALETTE_RGB24;		

	/* In case app doesn't set them... */
	for (i = 0; i < OVFX2_NUMFRAMES; i++) {
		ov->frame[i].width = ov->maxwidth;
		ov->frame[i].height = ov->maxheight;
		ov->frame[i].state = FREE_EMPTY;
		ov->frame[i].bytes_read = 0;
		ov->frame[i].depth = get_depth(ov->currentformat);
		PDEBUG(MISC|4, "%d: %dx%d depth %d", i, ov->frame[i].width, ov->frame[i].height, ov->frame[i].depth);
	}
		
	/* Initialize to max width/height, default palette */
	if (mode_init_regs(ov, ov->maxwidth, ov->maxheight, ov->currentformat, 0) < 0) {
		err = -EINVAL;
		goto out;
	}

	err = ovfx2_alloc(ov);
	if (err < 0)
		goto out;

	err = ovfx2_start_stream(ov);
	ov->curframe = -1;
		
	if (err) {
		ovfx2_dealloc(ov);
		ovfx2_dealloc_fbuf(ov);	/* don't forget fbuf (it won't have been MMAP-ed yet!) */
	}

out:
	up(&ov->lock);
	PDEBUG(FNCS|1, "Exiting ovfx2_initial_open()");
	return err;
}


void
ovfx2_free( struct usb_ovfx2 *ov )
{

	PDEBUG(FNCS|2, "Entering ov_free()");

	/*
	 * The next 2 lines are "belt & braces". The buffers should be free already
	 * because the device *should* be closed before we call ovfx2_free().
	 */		
	ovfx2_dealloc(ov);	/* let go of all the buffers and things */
	ovfx2_dealloc_fbuf(ov);	/* let go of 'fbuf' (can't be MMAP-ed by now) */
	
	/* The cbuf will still be allocated unless the USB was physically unplugged */
	ovfx2_cbuf_dealloc(ov);	/* let go of the command-buffer */

	/* let go the ov struct itself */
	down(&ov_free_lock);
	kfree(ov);
	up(&ov_free_lock);
	
	PDEBUG(MISC|1, "ov itself is free");
	PDEBUG(FNCS|2, "Exiting ov_free()");
	return;
}


/* Do not call this function directly! */
static int
ovfx2_do_ioctl( struct inode *inode, struct file *file, unsigned int cmd, void *arg )
{
	int	ret = -EINVAL;
//	struct video_device *vdev = file->private_data;
//	struct usb_ovfx2 *ov = video_get_drvdata(vdev);
	struct usb_ovfx2 *ov = ( struct usb_ovfx2 * ) file;

	switch (_IOC_TYPE(cmd)) {
	case 'v':
		PDEBUG(IOCTL|1, "ioctl 0x%x (v4l1, VIDIOC%s)", cmd,
		       (_IOC_NR(cmd) < NUM_V4L1_IOCTLS) ?
		       v4l1_ioctls[_IOC_NR(cmd)] : "???");
		break;
#if defined(HAVE_V4L2)
	case 'V':
		PDEBUG(IOCTL|1, "ioctl 0x%x (v4l2, %s)", cmd,
		       v4l2_ioctl_names[_IOC_NR(cmd)]);
		break;
#endif
	default:
		PDEBUG(IOCTL|1, "ioctl 0x%x (?)", cmd);
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -EIO;

	switch (cmd) {
	case VIDIOCGCAP:
	{
		struct video_capability *b = arg;

		memset(b, 0, sizeof(struct video_capability));
		sprintf(b->name, "OVFX2 USB Camera");
//		b->type = VID_TYPE_CAPTURE | VID_TYPE_SUBCAPTURE;
		b->type = VID_TYPE_CAPTURE;	/* let's *not* offer sub-windowing yet! */
		b->channels = 1;
		b->maxwidth = ov->maxwidth;
		b->maxheight = ov->maxheight;
		b->minwidth = ov->minwidth;
		b->minheight = ov->minheight;
		ret = 0;
		break;
	}
	case VIDIOCGCHAN:
	{
		struct video_channel *v = arg;

		if ((unsigned)(v->channel) < 2) {
			v->norm = 0;
			v->type = VIDEO_TYPE_CAMERA;
			v->flags = 0;
			v->tuners = 0;
			sprintf(v->name, "%s", "Camera");
			ret = 0;
		}
		
		break;
	}
	case VIDIOCSCHAN:
	{
		struct video_channel *v = arg;

		if (v->channel == 0) {
			/* standard mode */
//			ov->scaling = 0;
			ret = 0;
		}
		else if (v->channel == 1) {
			/* scaling mode: 800x600 -> 1600x1200 in the driver */
//			ov->scaling = 1;
			ret = 0;
		}
		
		break;
	}
	case VIDIOCGTUNER:
	{
		struct video_tuner *t = arg;
		
//		PDEBUG(IOCTL|2, "VIDIOCGTUNER: 0x%x, 0x%x, 0x%x (magic 0x%x)", t->tuner, t->mode, t->signal, t->flags);

		if (t->flags == ~(t->tuner ^ t->mode ^ t->signal) && t->tuner <= 0x3D) {
			/* magic sensor parameter twiddle */
			if (t->signal != 0) {
				PDEBUG(OUTGOING|2, "Twiddling register 0x%x to 0x%x - mask 0x%x", t->tuner, t->mode, t->signal);
				i2c_w_mask(ov, t->tuner, t->mode, t->signal);
			}
			
			/* read the register */
			t->mode = i2c_r(ov, t->tuner);
			ret = 0;
		}
		
		break;
	}
	case VIDIOCGPICT:
	{
		struct video_picture *p = arg;
		
		memset(p, 0, sizeof(struct video_picture));
# if 0
		if (sensor_get_picture(ov, p)) {
			PDEBUG(IOCTL|3, "Early bath");
			return -EIO;
		}
# else
		sensor_get_picture(ov, p);
# endif
		/* Can we get these from frame[0]? -claudio? */
		/* no you bloody can't!! - SH */
		/* we need to hold them in ov->depth/ov->currentformat and use them for read() grabs */
		/* no need to use them for MCAPTURE grabs as the MCAPTURE specifies them itself */
		p->depth = get_depth(ov->currentformat);
		p->palette = ov->currentformat;
		ret = 0;
		break;
	}
	case VIDIOCSPICT:
	{
		struct video_picture *p = arg;

		if (!get_depth(p->palette)) {
			PDEBUG(IOCTL|3, "Don't like depth");
		}
		else if (p->depth != get_depth(p->palette)) {
			PDEBUG(IOCTL|3, "Depth (%d) contradicts implied depth (%d)", p->depth, get_depth(p->palette));
		}
		else if (sensor_set_picture(ov, p)) {
			PDEBUG(IOCTL|3, "Don't like picture params");
			ret = -EIO;
		}
		else if (force_palette && p->palette != force_palette)
			info("SPICT: Palette rejected (%s)",
			     symbolic(v4l1_plist, p->palette));
		else {
			ov->currentformat = p->palette;
			ov->curframe = -1;
			ret = 0;
		}
		
		break;
	}
	case VIDIOCGCAPTURE:
	{
		int *vf = arg;

		ov->sub_flag = *vf;
		ret = 0;
		break;
	}
	case VIDIOCSCAPTURE:
	{
		struct video_capture *vc = arg;

		if (!vc->flags && !vc->decimation) {
			vc->x &= ~3L;
			vc->y &= ~1L;
			vc->y &= ~31L;

			if (vc->width == 0)
				vc->width = 32;

			vc->height /= 16;
			vc->height *= 16;
			
			if (vc->height == 0)
				vc->height = 16;

			ov->subx = vc->x;
			ov->suby = vc->y;
			ov->subw = vc->width;
			ov->subh = vc->height;
			ret = 0;
		}
		
		break;
	}
	case VIDIOCSWIN:
	{
		struct video_window *vw = arg;
		int i;

		PDEBUG(IOCTL|2, "Set window: %dx%d", vw->width, vw->height);

		if (vw->width > 800 || vw->height > 600) {
			/* use 1600x1200 mode */
			vw->width = 1600;
			vw->height = 1200;
		}
		else {
			/* use 800x600 mode */
			vw->width = 800;
			vw->height = 600;
		}

		PDEBUG(IOCTL|6, "Actually, set window: %dx%d", vw->width, vw->height);
# if 0
/* See earlier comments about foul-ups */

#if 0
/* This code breaks a few apps (Gnomemeeting, possibly
  others). Disable it until a solution can be found */
		if (vw->flags || vw->clipcount)
			return -EINVAL;
#endif

		rc = ovfx2_wait_frames_inactive(ov);
		if (rc)
			return rc;

		for (i = 0; i < OVFX2_NUMFRAMES; i++) {
			ov->frame[i].width = vw->width;
			ov->frame[i].height = vw->height;
		}
			
		if (ov->frame[0].width > 800 || ov->frame[0].height > 600) {
			ov->frame[0].reqdsensorstate = SENSOR_UXGA;
			ov->frame[1].reqdsensorstate = SENSOR_UXGA;
		}
		else {
			ov->frame[0].reqdsensorstate = SENSOR_SVGA;
			ov->frame[1].reqdsensorstate = SENSOR_SVGA;
		}
		
		PDEBUG(IOCTL|OUTGOING|3, "VIDIOSWIN sets sensor");
		rc = mode_reinit_regs(ov, &ov->frame[0], ov->sub_flag);
			
		if (rc < 0)
			return rc;
			
		ovfx2_resume(ov);
# else
		ov->curframe = -1;
		
		for (i = 0; i < OVFX2_NUMFRAMES; i++) {
			ov->frame[i].state = FRAME_ABORT;
			wake_up_interruptible(&ov->frame[i].wq);
		}
# endif
		ret = 0;
		break;
	}
	case VIDIOCGWIN:
	{
		struct video_window *vw = arg;

		memset(vw, 0, sizeof(struct video_window));
		vw->x = 0;		/* FIXME */
		vw->y = 0;
		vw->width = ov->frame[0].width;
		vw->height = ov->frame[0].height;
		vw->flags = 30;

		PDEBUG(IOCTL|2, "Get window: %dx%d", vw->width, vw->height);
		ret = 0;
		break;
	}
	case VIDIOCGMBUF:
	{
		struct video_mbuf *vm = arg;
		int i;

		memset(vm, 0, sizeof(struct video_mbuf));
		vm->size = OVFX2_NUMFRAMES
			   * MAX_DATA_SIZE(ov->maxwidth, ov->maxheight);
		vm->frames = OVFX2_NUMFRAMES;

		vm->offsets[0] = 0;
		for (i = 1; i < OVFX2_NUMFRAMES; i++) {
			vm->offsets[i] = vm->offsets[i-1]
			   + MAX_DATA_SIZE(ov->maxwidth, ov->maxheight);
		}

		ret = 0;
		break;
	}
	case VIDIOCMCAPTURE:
	{
		struct video_mmap *vm = arg;
		int depth;
		unsigned int f = vm->frame;
		struct ovfx2_frame *frame;

		PDEBUG(IOCTL|2, "MCAPTURE: frame: %d, %dx%d, %s", f, vm->width,
			vm->height, symbolic(v4l1_plist, vm->format));

		depth = get_depth(vm->format);
		if (!depth) {
			PDEBUG(IOCTL|3, "MCAPTURE: invalid format (%s)",
			       symbolic(v4l1_plist, vm->format));
		}
		else if (f >= OVFX2_NUMFRAMES)
			err("MCAPTURE: invalid frame (%d)", f);
		else if (vm->width > ov->maxwidth
		    || vm->height > ov->maxheight)
			err("MCAPTURE: requested dimensions too big");
		else {
			frame = &ov->frame[f];

			if (frame->state == MARKED_FILLING) {
				PDEBUG(IOCTL|4, "MCAPTURE: already grabbing");
				ret = -EBUSY;
			}
			else if (force_palette && (vm->format != force_palette)) {
				PDEBUG(IOCTL|3, "MCAPTURE: palette rejected (%s)",
			    	   symbolic(v4l1_plist, vm->format));
			}
			else {
				frame->width = vm->width;
				frame->height = vm->height;
				frame->sub_flag = ov->sub_flag;
				frame->depth = depth;
		
				if (ov->curframe == -1) {
					/* nothing actively capturing, so we can reset sensor now if we need to */
					PDEBUG(IOCTL|3, "MCAPTURE: resetting sensor");
					mode_reinit_regs(ov, frame, ov->sub_flag);
				}
		
				/* Mark it as ready for filling */
				ret = ovfx2_new_frame(ov, f, vm->format);
			}
		}
		
		break;
	}
	case VIDIOCSYNC:
	{
		unsigned int fnum = *((unsigned int *) arg);
		struct ovfx2_frame *frame;
		int rc;

		if (fnum >= OVFX2_NUMFRAMES) {
			err("SYNC: invalid frame (%d)", fnum);
			return -EINVAL;
		}

		frame = &ov->frame[fnum];

		PDEBUG(IOCTL|FRAMESTATE|2, "syncing to frame %d, state = %d", fnum,
		       frame->state);

		switch (frame->state) {
		case FRAME_ABORT:
			frame->state = FREE_EMPTY;
			/* fall thru */
			
		case FREE_EMPTY:
		case FREE_FILLING:
		case FREE_FULL:
			/* no MCAPTURE was ever issued on this frame */
			return -EINVAL;
			
		case MARKED_EMPTY:
		case MARKED_FILLING:
		case FRAME_ERROR:
redo:
			up(&ov->lock);					/* let the semaphore go whilst waiting: device might get unplugged */
			PDEBUG(FRAMESTATE|3, "waiting on frame %d, state = %d", fnum, frame->state);
			rc = wait_event_interruptible(frame->wq,
			    ov->fx2state != FX2_STREAMING
			    || (frame->state == MARKED_FULL)
			    || (frame->state == FRAME_ERROR)
			    || (frame->state == FRAME_ABORT)
			    || ov->rawfbuf == NULL);

			PDEBUG(FRAMESTATE|3, "finished waiting on frame %d, state = %d", fnum, frame->state);
			
			if (down_interruptible(&ov->lock)) {
				/* failure here might be a bit fatal: ovfx2_ioctl() above us is going to try and 'up' this semaphore when we return */
				PDEBUG(FNCS|1, "ovfx2_do_ioctl(): can't regain semaphore");
				return -EINTR;
			}

			if (rc)
				return rc;

			if (ov->rawfbuf == NULL)
				/* looks like the device went away */
				return -ENODEV;
			
			if (ov->fx2state == FX2_WEDGED /* || ov->fx2state == FX2_WRONGSIZE */) {
				PDEBUG(BOMB|FRAMESTATE|OUTGOING|1, "SYNC: the %s went off", (ov->fx2state == FX2_WRONGSIZE)? "wrong size event": "bomb");
				PDEBUG(BOMB|1, "BOMB: last known actlen: %d", last_actlen);
					
				ovfx2_stop_stream(ov);
				ov->sensorstate = SENSOR_UNKNOWN_MODE;
								
				if (ov->curframe >= 0) {
					if (mode_init_regs(ov, frame->width, frame->height, ov->currentformat, 0) < 0)
						return -EINVAL;
				}
				
				ovfx2_start_stream(ov);
				frame->state = FRAME_ERROR;
			}
			else if (ov->fx2state == FX2_WRONGMODE || ov->fx2state == FX2_WRONGSIZE) {
				ovfx2_stop_stream(ov);
				
				if (ov->curframe >= 0) {
					mode_reinit_regs(ov, &ov->frame[ov->curframe], ov->sub_flag);
				}
				
				ovfx2_start_stream(ov);
				frame->state = FRAME_ERROR;
			}
				
			if (frame->state == FRAME_ABORT) {
				PDEBUG(FRAMESTATE|2, "SYNC: returning error for frame %d", fnum);
				frame->state = FREE_EMPTY;
				return -EINVAL;
			}
			else if (frame->state == FRAME_ERROR) {
				if ((rc = ovfx2_new_frame(ov, fnum, -1)) < 0)
					return rc;
			}
			
			if (frame->state != MARKED_FULL)
				goto redo;
				
			/* Fall through */
			
		case MARKED_FULL:
			PDEBUG(FRAMESTATE|4, "SYNC: frame %d is DONE", fnum);

			/* Decompression, format conversion, etc... */
			ovfx2_postprocess(ov, frame);

			frame->state = FREE_EMPTY;
			PDEBUG(FRAMESTATE|4, "SYNC: frame %d now FREE_EMPTY", fnum);
			
			if (ov->curframe == -1) {
				PDEBUG(FRAMESTATE|4, "SYNC: starting opportunistic filling of %d", fnum);
				ov->curframe = fnum;
			}
			
			break;
		} /* end switch */

		return 0;
	}
	case VIDIOCGFBUF:
	{
		struct video_buffer *vb = arg;

		memset(vb, 0, sizeof(struct video_buffer));

		return 0;
	}
	case VIDIOCGUNIT:
	{
		struct video_unit *vu = arg;

		memset(vu, 0, sizeof(struct video_unit));

		vu->video = ov->minor;
		vu->vbi = VIDEO_NO_UNIT;
		vu->radio = VIDEO_NO_UNIT;
		vu->audio = VIDEO_NO_UNIT;
		vu->teletext = VIDEO_NO_UNIT;
		ret = 0;
		break;
	}
	default:
		PDEBUG(IOCTL|1, "Unsupported IOCtl: 0x%X", cmd);
		ret = -ENOIOCTLCMD;	/* V4L layer handles this */
	} /* end switch */

	PDEBUG(IOCTL|1, "ioctl returns: %d", ret);
	return ret;
}


int
ovfx2_ioctl(struct usb_ovfx2 *ov, struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
	int rc;

	if (!ov->rawfbuf) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_ioctl(): no device");
		return -EIO;
	}
	
	if (down_interruptible(&ov->lock)) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_ioctl(): semaphore");
		return -EINTR;
	}

	PDEBUG(FNCS|1, "Entering ovfx2_ioctl()");
	/* misuse of 2nd param below... */
	rc = video_usercopy(inode, ( struct file * )ov, cmd, arg, ovfx2_do_ioctl);
	PDEBUG(FNCS|1, "Exiting ovfx2_ioctl()");
	
	up(&ov->lock);
	return rc;
}


int
ovfx2_read( struct usb_ovfx2 *ov, struct file *file, char *buf, size_t cnt, loff_t *ppos )
{
	int noblock = file->f_flags&O_NONBLOCK;
	unsigned long count = cnt;
	int i, rc = 0, frmx = -1;
	struct ovfx2_frame *frame;

	if (down_interruptible(&ov->lock)) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_read(): semaphore");
		return -EINTR;
	}

	PDEBUG(MISC|4, "%ld bytes, noblock=%d", count, noblock);

	if (!buf) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_read(): fault#2");
		rc = -EFAULT;
		goto error;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf) || ov->rawfbuf == NULL) {
#else
	if (!ov->dev || ov->rawfbuf == NULL) {
#endif
		PDEBUG(FNCS|1, "Repelled from ovfx2_read(): no device");
		rc = -ENODEV;
		goto error;
	}

// FIXME: Only supports two frames
	/* See if a frame is completed, then use it. */
	if (ov->frame[0].state == MARKED_FULL || ov->frame[0].state == FRAME_ERROR)
		frmx = 0;
	else if (ov->frame[1].state >= MARKED_FULL || ov->frame[1].state == FRAME_ERROR)
		frmx = 1;

	/* If nonblocking we return immediately */
	if (noblock && (frmx == -1)) {
		rc = -EAGAIN;
		goto error;
	}

	/* If no MARKED_FULL, look for a MARKED_FILLING state. */
	/* See if a frame is in process (grabbing or filling), then use it. */
	if (frmx == -1) {
		if (ov->frame[0].state == MARKED_FILLING || ov->frame[0].state == FREE_FILLING)
			frmx = 0;
		else if (ov->frame[1].state == MARKED_FILLING || ov->frame[1].state == FREE_FILLING)
			frmx = 1;
	}

	/* If no frame is active, start one. */
	if (frmx == -1) {
		if ((rc = ovfx2_new_frame(ov, frmx = 0, ov->currentformat))) {
			err("read: ovfx2_new_frame error");
			goto error;
		}
	}

	frame = &ov->frame[frmx];

restart:
	/* Wait while we're grabbing the image */
	PDEBUG(MISC|4, "Waiting image grabbing");
	up(&ov->lock);					/* let the semaphore go whilst waiting: device might get unplugged */
	rc = wait_event_interruptible(frame->wq,
		(frame->state == MARKED_FULL)
		|| (frame->state == FRAME_ERROR)
		|| ov->rawfbuf == NULL);

	if (rc)
		return rc;
		
	if (down_interruptible(&ov->lock)) {
		PDEBUG(FNCS|1, "ovfx2_read(): Can't regain semaphore");
		return -EINTR;
	}

	if (ov->rawfbuf == NULL) {
		rc = -ENODEV;
		goto error;
	}

	PDEBUG(MISC|4, "Got image, frame->state = %d", frame->state);
	PDEBUG(MISC|4, "bytes_recvd = %d", frame->bytes_recvd);

	if (frame->state == FRAME_ERROR) {
		frame->bytes_read = 0;
		err("** ick! ** Errored frame %d", ov->curframe);
		if (ovfx2_new_frame(ov, frmx, ov->currentformat)) {
			err("read: ovfx2_new_frame error");
			goto error;
		}
		goto restart;
	}

	/* Decompression, format conversion, etc... */
	ovfx2_postprocess(ov, frame);

	PDEBUG(MISC|4, "frmx=%d, bytes_read=%ld, length=%ld", frmx,
		frame->bytes_read,
		get_frame_length(frame));

	/* copy bytes to user space; we allow for partials reads */
//	if ((count + frame->bytes_read)
//	    > get_frame_length((struct ovfx2_frame *)frame))
//		count = frame->scanlength - frame->bytes_read;

	/* FIXME - count hardwired to be one frame... */
	count = get_frame_length(frame);

	PDEBUG(MISC|4, "Copy to user space: %ld bytes", count);
	if ((i = copy_to_user(buf, frame->data + frame->bytes_read, count))) {
		PDEBUG(MISC|4, "Copy failed! %d bytes not copied", i);
		rc = -EFAULT;
		goto error;
	}

	frame->bytes_read += count;
	PDEBUG(MISC|4, "{copy} count used=%ld, new bytes_read=%ld",
		count, frame->bytes_read);

	/* If all data have been read... */
	if (frame->bytes_read
	    >= get_frame_length(frame)) {
		frame->bytes_read = 0;

// FIXME: Only supports two frames
		/* Mark it as available to be used again. */
		ov->frame[frmx].state = FREE_EMPTY;
		if ((rc = ovfx2_new_frame(ov, !frmx, ov->currentformat))) {
			err("ovfx2_new_frame returned error");
			goto error;
		}
	}

	PDEBUG(MISC|4, "read finished, returning %ld (sweet)", count);

	up(&ov->lock);
	return count;

error:
	up(&ov->lock);
	return rc;
}


int
ovfx2_mmap( struct usb_ovfx2 *ov, struct file *file, struct vm_area_struct *vma )
{
	unsigned long start = vma->vm_start;
	unsigned long size  = vma->vm_end - vma->vm_start;
	unsigned long page, pos;

	if (down_interruptible(&ov->lock)) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_mmap(): semaphore");
		return -EINTR;
	}
		
	if (size > (((OVFX2_NUMFRAMES
	              * MAX_DATA_SIZE(ov->maxwidth, ov->maxheight)
	              + PAGE_SIZE - 1) & ~(PAGE_SIZE - 1)))) {
		up(&ov->lock);
		PDEBUG(FNCS|1, "Repelled from ovfx2_mmap(): bad params");
		return -EINVAL;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf) || ov->rawfbuf == NULL) {
#else
	if (!ov->dev || ov->rawfbuf == NULL) {
#endif
		up(&ov->lock);
		PDEBUG(FNCS|1, "Repelled from ovfx2_mmap(): no dev");
		return -ENODEV;
	}

	PDEBUG(FNCS|1, "Entering ovfx2_mmap()");
	PDEBUG(MISC|4, "mmap: %ld (%lX) bytes", size, size);
	
	pos = (unsigned long)ov->fbuf;
	while (size > 0) {
		page = kvirt_to_pa(pos);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 3) || defined(RH9_REMAP)
# if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 10)
		if (remap_pfn_range(vma, start, page >> PAGE_SHIFT, PAGE_SIZE, PAGE_SHARED)) {
# else
		if (remap_page_range(vma, start, page, PAGE_SIZE, PAGE_SHARED)) {
# endif
#else
		if (remap_page_range(start, page, PAGE_SIZE, PAGE_SHARED)) {
#endif
			up(&ov->lock);
			PDEBUG(FNCS|1, "Exiting ovfx2_mmap(): can't remap");
			return -EAGAIN;
		}
		start += PAGE_SIZE;
		pos += PAGE_SIZE;
		if (size > PAGE_SIZE)
			size -= PAGE_SIZE;
		else
			size = 0;
	}

	up(&ov->lock);
	PDEBUG(FNCS|1, "Exiting ovfx2_mmap()");
	return 0;
}


/****************************************************************************
 *
 * FX2 and sensor configuration
 *
 ***************************************************************************/

/* This is called when an OV2610 */
static int
ov2xx0_configure(struct usb_ovfx2 *ov)
{

	if (ov->sensor != CC_OV2610) // CC_OV7620) FIXME-JSM 
	{
		err ("Wrong cam chip detected %X\n", ov->sensor);
		return -ENODEV;
	}

	ov->internal_client.addr =  OV2xx0_SID; //OV7xx0_SID;

	ov->maxwidth = 1600; //640;
	ov->maxheight = 1200; //480;

	ov->minwidth = 800;
	ov->minheight = 600;

	return 0; //camchip_init_settings(ov);
}

/* This is called when an OV7610, OV7620, or OV76BE is detected. */
static int
ov7xx0_configure(struct usb_ovfx2 *ov)
{
	if (ov->sensor != CC_OV7620)
	{
		return -ENODEV;
	}

	ov->internal_client.addr =  OV7xx0_SID;

	ov->maxwidth = 640;
	ov->maxheight = 480;

#if 0
	ov->minwidth = 64;
	ov->minheight = 48;
#else
	/* Can't do less than max res yet */
	ov->minwidth = ov->maxwidth;
	ov->minheight = ov->maxheight;
#endif

	return camchip_init_settings(ov);
}

/* This initializes the FX2 and the sensor */
static int
ovfx2_configure(struct usb_ovfx2 *ov)
{
	static struct regval regvals_init[] = {
		/* Reset FX2 */
		{ 0x0f, 0x01, 0xff },
		{ 0x0f, 0x03, 0xff },
		{ 0x0f, 0x01, 0xff },
# if HARDCODED_SENSOR
		{ 0x0f, 0x08, 0x08 },
		
		/* Default camera interface settings */
		{ 0xee, 0x80, 0xff },
		{ 0x0f, 0x04, 0x04 },
		{ 0xe0, 0x46, 0xff },
		{ 0xe1, 0xf8, 0xff },
		{ 0xe2, 0x02, 0xff },
		{ 0xe3, 0x07, 0xff },
		{ 0xe4, 0x49, 0xff },
		{ 0xe5, 0xf0, 0xff },
		{ 0xe6, 0x0c, 0xff },
		{ 0xe7, 0xf4, 0xff },
		{ 0xe8, 0x3f, 0xff },

		{ 0x0f, 0x80, 0x80 },

		/* Default picture settings */
		{ REG_SAT,	 187, 0xff },
		{ REG_HUE,	0x00, 0xff },
		{ REG_CNTR,	0x80, 0xff },
		{ REG_BRIGHT,	0x14, 0xff },
		{ REG_SHARP,	  88, 0xff },
# else
		/* Cribbed from Omnivision's ov2610.set file */
		{ 0x02, FIFO8, 0xff },		/* set control-bit 0 (FIFO width=8) */
		{ 0xe0, 0x51, 0xff },
		{ 0xe1, 0xfa, 0xff },
		{ 0xe2, 0xf5, 0xff },
		{ 0xe3, 0xc9, 0xff },
		{ 0xe4, 0x74, 0xff },
		{ 0xe5, 0x03, 0xff }, 
		{ 0xe6, 0xfa, 0xff }, 
		{ 0xe7, 0xf8, 0xff },
		{ 0xe8, 0x4e, 0xff },
		{ 0xe9, 0x80, 0xff },		/* saturation? */
		{ 0xea, 0x00, 0xff },		/* hue? */
//		{ 0xec, 0x00, 0xff },		/* brightness? */
		{ 0xed, 0x01, 0xff },		/* gamma? [range 0 -> 1] */
		{ 0xee, 0x80, 0xff },		/* sharpness? [range 0 -> 255] */
		{ 0xef, 0x01, 0xff },
		{ 0xf1, 0x02, 0xff },
		{ 0x0f, STARTUNK, UNKMASK },		/* set control-bit 2 (release from RESET maybe?) */
		{ 0xef, 0x00, 0xff },
		{ 0xed, 0x00, 0xff },
		{ 0xf1, 0x00, 0xff },
		{ 0xfe, 0x01, 0xff },
# endif
		{ 0x00, 0x00, 0x00 }, /* end */
	};

	ov->frame[0].bayer_decode = bayer_decode_BGBG_GRGR_8;
	ov->frame[0].rematrix_even = ov->frame[0].rematrix_odd = Cam2RGB24;
	ov->frame[1].bayer_decode = bayer_decode_BGBG_GRGR_8;
	ov->frame[1].rematrix_even = ov->frame[1].rematrix_odd = Cam2RGB24;

	ovfx2_init_i2c(ov);
	
	if (write_regvals(ov, regvals_init) == 0)
		return 0;

	err("OVFX2 Config failed");
	return -EBUSY;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
/****************************************************************************
 *  sysfs
 ***************************************************************************/
# if 0
static inline struct usb_ovfx2 *cd_to_ov(struct class_device *cd)
{
	struct video_device *vdev = to_video_device(cd);
	return video_get_drvdata(vdev);
}

static ssize_t show_model(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	return sprintf(buf, "%s\n", ov->desc);
} 
static CLASS_DEVICE_ATTR(model, S_IRUGO, show_model, NULL);


static ssize_t show_sensor(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	return sprintf(buf, "%s\n", symbolic(senlist, ov->sensor));
} 
static CLASS_DEVICE_ATTR(sensor, S_IRUGO, show_sensor, NULL);

static ssize_t show_brightness(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	int x;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -ENODEV;
		
	sensor_get_control(ov, OVCAMCHIP_CID_BRIGHT, &x);
	return sprintf(buf, "%d\n", x);
} 
static CLASS_DEVICE_ATTR(brightness, S_IRUGO, show_brightness, NULL);

static ssize_t show_saturation(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	int x;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -ENODEV;
		
	sensor_get_control(ov, OVCAMCHIP_CID_SAT, &x);
	return sprintf(buf, "%d\n", x);
} 
static CLASS_DEVICE_ATTR(saturation, S_IRUGO, show_saturation, NULL);

static ssize_t show_contrast(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	int x;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -ENODEV;
		
	sensor_get_control(ov, OVCAMCHIP_CID_CONT, &x);
	return sprintf(buf, "%d\n", x);
} 
static CLASS_DEVICE_ATTR(contrast, S_IRUGO, show_contrast, NULL);

static ssize_t show_hue(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	int x;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -ENODEV;
		
	sensor_get_control(ov, OVCAMCHIP_CID_HUE, &x);
	return sprintf(buf, "%d\n", x);
} 
static CLASS_DEVICE_ATTR(hue, S_IRUGO, show_hue, NULL);

static ssize_t show_exposure(struct class_device *cd, char *buf)
{
	struct usb_ovfx2 *ov = cd_to_ov(cd);
	int x;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!ov->intf || !interface_to_usbdev(ov->intf))
#else
	if (!ov->dev)
#endif
		return -ENODEV;
		
	sensor_get_control(ov, OVCAMCHIP_CID_EXP, &x);
	return sprintf(buf, "%d\n", x);
} 
static CLASS_DEVICE_ATTR(exposure, S_IRUGO, show_exposure, NULL);

static void ov_create_sysfs(struct video_device *vdev)
{
	video_device_create_file(vdev, &class_device_attr_model);
	video_device_create_file(vdev, &class_device_attr_sensor);
	video_device_create_file(vdev, &class_device_attr_brightness);
	video_device_create_file(vdev, &class_device_attr_saturation);
	video_device_create_file(vdev, &class_device_attr_contrast);
	video_device_create_file(vdev, &class_device_attr_hue);
	video_device_create_file(vdev, &class_device_attr_exposure);
}
#endif
#endif

/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/


/*
 * We come here as a result of either:
 * 1) Issuing an "insmod" on the driver with the camera plugged in at the time.
 * 2) The camera being plugged in whilst the driver is loaded.
 *
 * It seems not to be possible to tell the difference between these two
 * instances.
 */
struct usb_ovfx2 *
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
ovfx2_probe( struct usb_interface *intf, const struct usb_device_id *id, int minor )
{
#else
ovfx2_probe( struct usb_device *dev, const struct usb_device_id *id, int minor )
{
#endif
	struct usb_ovfx2 *ov;
	int i;

	PDEBUG(FNCS|1, "Entering ovfx2_probe()");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;
#endif

	if ((ov = kmalloc(sizeof(*ov), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc ov struct");
		goto error_out;
	}

	memset(ov, 0, sizeof(*ov));
	PDEBUG(MISC|1, "ov itself allocated");

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	ov->intf = intf;
#else
	ov->dev = dev;
#endif

	ov->stop_during_set = !fastset;
	ov->sensor = CC_UNKNOWN;
	ov->bomb_running = 0;
	ov->minor = minor;

	/* FIXME: Determine this from the VID/PID */
	ov->bridge = BRG_2800;		/* iBot2 */

	init_waitqueue_head(&ov->wq);

	init_MUTEX(&ov->lock);	/* to 1 == available */
	init_MUTEX(&ov->buf_lock);
	init_MUTEX(&ov->cbuf_lock);

	ov->buf_state = BUF_NOT_ALLOCATED;

	/* Allocate control transfer buffer. */
	/* Must be kmalloc()'ed, for DMA compatibility */
	ov->cbuf = kmalloc(OVFX2_CBUF_SIZE, GFP_KERNEL);
	if (!ov->cbuf)
		goto error;
		
	PDEBUG(MISC|1, "ov->cbuf allocated");

	if (ovfx2_configure(ov) < 0)
		goto error;

	for (i = 0; i < OVFX2_NUMFRAMES; i++) {
		ov->frame[i].framenum = i;
		init_waitqueue_head(&ov->frame[i].wq);
	}

	for (i = 0; i < OVFX2_NUMSBUF; i++) {
		ov->sbuf[i].ov = ov;
		spin_lock_init(&ov->sbuf[i].lock);
		ov->sbuf[i].n = i;
	}

#ifdef MODULE_DEBUG
	if (0 && dump_bridge)
		ovfx2_dump_regs(ov);
#endif
	sprintf(i2c_adapname(&ov->i2c_adap), "OVFX2 #%d", ov->minor);

	create_proc_ovfx2_cam(ov);

	info("FX2 probe complete");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	PDEBUG(FNCS|1, "Exiting ovfx2_probe()");
	return ov;

error:
	ovfx2_free(ov);
	ov = NULL;

error_out:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	err("Camera initialization failed");
	PDEBUG(FNCS|1, "Exiting ovfx2_probe(): failed");
	return NULL;
}


int
ovfx2_stop_and_release( struct usb_ovfx2 *ov )
{
	int	n;
	
	if (!ov) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_stop_and_release(): no device");
		return -ENODEV;
	}

	PDEBUG(FNCS|1, "Entering ovfx2_stop_and_release()");
	
	/*
	 * We may have been in the middle of streaming.
	 * Stop streaming.
	 */
	ovfx2_stop_stream(ov);	/* unlinks the URBs too */
	
	PDEBUG(MISC|3, "Stream stopped");

	for (n = 0; n < OVFX2_NUMFRAMES; n++)
		ov->frame[n].state = FRAME_ERROR;

	ov->curframe = -1;
	
# if 0 /* ovfx2_dealloc() does this (and more) for us properly */
	ov->rawfbuf = NULL;
	
	/*
	 * This will release the errored frames back to userland. Doing so
	 * will probably cause the user process to re-issue the request(s).
	 * However, those IOCTLs ought to bomb out with -ENODEV because ov->rawfbuf
	 * is now NULL.
	 *
	 * In fact, all calls (except close()) will bomb out with -ENODEV...
	 */
	for (n = 0; n < OVFX2_NUMFRAMES; n++)
		wake_up_interruptible(&ov->frame[n].wq);

	wake_up_interruptible(&ov->wq);
	ov->buf_state = BUF_NOT_ALLOCATED;	/* this is a bit silly, could just check the pointers */
# else
	ovfx2_dealloc(ov);	/* free the frame's data buffers, but not 'fbuf'. ( Releases any pending CSYNC ioctl() or read() ) */
# endif

	/* Can't take lock earlier since it would deadlock
	 * read() or ioctl() if open and sleeping */
	down(&ov->lock);

	ov->fx2state = FX2_NOT_STREAMING;
	ov->fsync = 0;
	ov->currentformat = (force_palette)? force_palette: VIDEO_PALETTE_RGB24;
	ov->sensorstate = SENSOR_UNKNOWN_MODE;
	up(&ov->lock);
	
	PDEBUG(FNCS|1, "Exiting ovfx2_stop_and_release()");
	return 0;
}


int
ovfx2_final_close( struct usb_ovfx2 *ov )
{

	ovfx2_stop_and_release(ov);
	ovfx2_dealloc_fbuf(ov);	/* we can only release the fbuf on final close */
	return 0;
}


/*
 * We come here as a result of the camera being unplugged.
 */
void
ovfx2_disconnect( struct usb_ovfx2 *ov )
{
	
	if (!ov) {
		PDEBUG(FNCS|1, "Repelled from ovfx2_disconnect(): no device");
		return;
	}
	
	PDEBUG(FNCS|1, "Entering ovfx2_disconnect()");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_INC_USE_COUNT;
#endif

	/* 
	 * Release the command-buffer.
	 * With the device unplugged, we can't (and mustn't) talk to it anyway!
	 */
	ovfx2_cbuf_dealloc(ov);

# if 0 /* pretty sure we shouldn't fiddle with 'dev' or with references to it... SH */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	dev_set_drvdata(&intf->dev, NULL);
	ov->intf = NULL;
#else
	ov->dev = NULL;
#endif
# endif

	ovfx2_stop_and_release(ov); /* stop streaming, release the URBs, release the data buffers BUT **NOT** FBUF */

	down(&ov->lock);
        destroy_proc_ovfx2_cam(ov);
	i2c_del_adapter(&ov->i2c_adap);
	up(&ov->lock);
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	PDEBUG(FNCS|1, "Exiting ovfx2_disconnect()");
	return;
}


/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

/*
 * We come here as a consequence of loading the driver with '/sbin/insmod'
 */ 
static int __init
ovfx2_init(void)
{

	PDEBUG(FNCS|1, "Entering ovfx2_init()");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	EXPORT_NO_SYMBOLS;
#endif
	init_MUTEX(&ov_free_lock);

        proc_ovfx2_create();

	info(DRIVER_VERSION " : " DRIVER_DESC);
	
# ifdef MODULE_DEBUG
	if ((debug & ~LMASK) == 0)
		/* in case user forgot to specify */
		debug |= MISC;
# endif /* MODULE_DEBUG */
	
	PDEBUG(FNCS|1, "Exiting ovfx2_init()");
	return 0;
}


/*
 * We come here as a consequence of unloading the driver with '/sbin/rmmod'
 */ 
static void __exit
ovfx2_exit(void)
{

	PDEBUG(FNCS|1, "Entering ovfx2_exit()");
        proc_ovfx2_destroy();
	PDEBUG(FNCS|1, "Exiting ovfx2_exit()");
	return;
}

module_init(ovfx2_init);
module_exit(ovfx2_exit);
