/*
 * Outer layer of a driver for XL Imaging's 2.1 Megapixel USB Camera.
 * (Uses the Omnivision OV2610 sensor chips and Omnivision's "FX2-9000" USB bridge
 * chip).
 *
 * It will also drive Omnivision's "2610-ECX" demo board.
 *
 * Copyright (c) 2004 Steve Hosgood <steve@caederus.com>
 * Copyright (c) 2004 XL Imaging Ltd, Singleton Park, Swansea SA2 8PP, United Kingdom.
 *
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version. NO WARRANTY OF ANY KIND is expressed or implied.
 */
 
# include	<linux/config.h>
# include	<linux/version.h>
# include	<linux/module.h>
# include	<linux/kernel.h>

/*
 * Version Information
 */
# define DESC		"XL Imaging 2.1Mpixel USB Camera driver"
# define VERSION	"v0.99"
# define EMAIL		"steve@xlimaging.com"
# define AUTHOR		"Steve Hosgood <steve@xlimaging.com>"
# define COPYRIGHT	"2004,2005, XL Imaging Ltd, Swansea SA2 8PP, United Kingdom."

# define XLI_2M_MAX_UNIT_VIDEO	16

static int unit_video[XLI_2M_MAX_UNIT_VIDEO];

# define MODULE_DEBUG
# include	"ovdebug.h"

# ifdef MODULE_DEBUG
static int debug;

MODULE_PARM(debug, "i");
MODULE_PARM_DESC(debug,
  "Debug level: 0=none, 1=inits, 2=warning, 3=config, 4=functions, 5=max");
# endif
MODULE_PARM(unit_video, "1-" __MODULE_STRING(XLI_2M_MAX_UNIT_VIDEO) "i");
MODULE_PARM_DESC(unit_video,
  "Force use of specific minor number(s). 0 is not allowed.");

MODULE_AUTHOR(AUTHOR);
MODULE_DESCRIPTION(DESC);
#if defined(MODULE_LICENSE)	/* Introduced in ~2.4.10 */
MODULE_LICENSE("GPL");
#endif


/* Prevents double-free of xli struct */
static struct semaphore xli_free_lock;

# include	"xli_2M.h"


/**********************************************************************
 * /proc interface
 * Based on the CPiA driver version 0.7.4 -claudio
 **********************************************************************/

#if defined(CONFIG_VIDEO_PROC_FS)

static struct proc_dir_entry *xli_2M_proc_entry = NULL;
extern struct proc_dir_entry *video_proc_entry;

/* Prototypes */
static int sensor_get_picture(struct usb_xli_2M *, struct video_picture *);
static int sensor_get_control(struct usb_xli_2M *, int, int *);

#define YES_NO(x) ((x) ? "yes" : "no")

/* /proc/video/xli_2M/<minor#>/info */
static int
xli_2M_read_proc_info(char *page, char **start, off_t off, int count, int *eof,
		     void *data)
{
	char *out = page;
	int i, len;
	struct usb_xli_2M *ov = data;
	struct video_picture p;
	int exp = 0;

	if (!ov || !ov->dev)
		return -ENODEV;

	sensor_get_picture(ov, &p);
	sensor_get_control(ov, OVCAMCHIP_CID_EXP, &exp);

	/* IMPORTANT: This output MUST be kept under PAGE_SIZE
	 *            or we need to get more sophisticated. */

	out += sprintf(out, "driver_version  : %s\n", DRIVER_VERSION);
	out += sprintf(out, "model           : %s\n", ov->desc);
	out += sprintf(out, "streaming       : %s\n", YES_NO(ov->streaming));
	out += sprintf(out, "grabbing        : %s\n", YES_NO(ov->grabbing));
	out += sprintf(out, "subcapture      : %s\n", YES_NO(ov->sub_flag));
	out += sprintf(out, "sub_size        : %d %d %d %d\n",
		       ov->subx, ov->suby, ov->subw, ov->subh);
	out += sprintf(out, "brightness      : %d\n", p.brightness >> 8);
	out += sprintf(out, "colour          : %d\n", p.colour >> 8);
	out += sprintf(out, "contrast        : %d\n", p.contrast >> 8);
	out += sprintf(out, "hue             : %d\n", p.hue >> 8);
	out += sprintf(out, "exposure        : %d\n", exp);
	out += sprintf(out, "num_frames      : %d\n", xli_2M_NUMFRAMES);
	for (i = 0; i < xli_2M_NUMFRAMES; i++) {
		out += sprintf(out, "frame           : %d\n", i);
		out += sprintf(out, "  size          : %d %d\n",
			       ov->frame[i].width, ov->frame[i].height);
		out += sprintf(out, "  format        : %s\n",
			       symbolic(v4l1_plist, ov->frame[i].format));
	}
	out += sprintf(out, "sensor          : %s\n",
		       symbolic(senlist, ov->sensor));
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 20)
	out += sprintf(out, "topology        : %s\n", ov->usb_path);
#else
	out += sprintf(out, "usb_bus         : %d\n", ov->dev->bus->busnum);
	out += sprintf(out, "usb_device      : %d\n", ov->dev->devnum);
#endif
	out += sprintf(out, "i2c_adap_id     : %d\n",
	               i2c_adapter_id(&ov->i2c_adap));

	len = out - page;
	len -= off;
	if (len < count) {
		*eof = 1;
		if (len <= 0)
			return 0;
	} else
		len = count;

	*start = page + off;

	return len;
}

static void
create_proc_xli_2M_cam(struct usb_xli_2M *ov)
{
	char dirname[10];

	if (!xli_2M_proc_entry || !ov) {
		info("No /proc entry to add to");
		return;
	}else
		info("Adding a /proc entry");

	/* Create per-device directory */
	snprintf(dirname, 10, "%d", ov->vdev->minor);
	ov->proc_devdir = create_proc_entry(dirname, S_IFDIR, xli_2M_proc_entry);
	if (!ov->proc_devdir)
		return;
	ov->proc_devdir->owner = THIS_MODULE;

	/* Create "info" entry (human readable device information) */
	ov->proc_info = create_proc_read_entry("info", S_IFREG|S_IRUGO|S_IWUSR,
		ov->proc_devdir, xli_2M_read_proc_info, ov);
	if (!ov->proc_info)
		return;
	ov->proc_info->owner = THIS_MODULE;
}

static void
destroy_proc_xli_2M_cam(struct usb_xli_2M *ov)
{
	char dirname[10];

	if (!ov || !ov->proc_devdir)
		return;

	snprintf(dirname, 10, "%d", ov->vdev->minor);

	/* Destroy "info" entry */
	if (ov->proc_info) {
		remove_proc_entry("info", ov->proc_devdir);
		ov->proc_info = NULL;
	}

	/* Destroy per-device directory */
	remove_proc_entry(dirname, xli_2M_proc_entry);
	ov->proc_devdir = NULL;
}

static void
proc_xli_2M_create(void)
{
	if (video_proc_entry == NULL) {
		err("Error: /proc/video/ does not exist");
		return;
	}

	xli_2M_proc_entry = create_proc_entry("xli_2M", S_IFDIR,
					     video_proc_entry);
	info("Created proc entry");

	if (xli_2M_proc_entry)
		xli_2M_proc_entry->owner = THIS_MODULE;
	else
		err("Unable to create /proc/video/xli_2M");
}

static void
proc_xli_2M_destroy(void)
{
	if (xli_2M_proc_entry == NULL)
		return;

	remove_proc_entry("xli_2M", video_proc_entry);
}
#else
static inline void create_proc_xli_2M_cam(struct usb_xli_2M *ov) { }
static inline void destroy_proc_xli_2M_cam(struct usb_xli_2M *ov) { }
static inline void proc_xli_2M_create(void) { }
static inline void proc_xli_2M_destroy(void) { }
#endif /* #ifdef CONFIG_VIDEO_PROC_FS */


static void
xli_2M_unregister( struct usb_xli_2M *xli )
{

	PDEBUG(FNCS|2, "Entering xli_2M_unregister()");
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
	if (xli->vdev) {
		video_device_release(xli->vdev);
		PDEBUG(MISC|2, "unregistered from V4L layer");
	}
#else
	if (xli->vdev) {
		video_unregister_device(xli->vdev);
		video_set_drvdata(xli->vdev, NULL);		/* make sure 'vdev' forgets the address of this 'xli' struct */
//		video_device_release(xli->vdev);		/* a.k.a   kfree(xli->vdev): previous line is a bit pointless then.. SH */
		PDEBUG(MISC|2, "unregistered from V4L layer");
	}
			
	/* interface must forget about the xli struct if we're freeing it! */
	usb_set_intfdata(xli->intf, NULL);
#endif
	xli->vdev = NULL;
	PDEBUG(FNCS|2, "Exiting xli_2M_unregister()");
	return;
}



/****************************************************************************
 *
 * V4L API
 *
 ***************************************************************************/

static int
xli_2M_open(struct inode *inode, struct file *file)
{
	struct video_device *vdev = video_devdata(file);
	struct usb_xli_2M *xli = video_get_drvdata(vdev);
	int err;

	if (xli == NULL) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_open(): no xli struct");
		return -EFAULT;
	}

	PDEBUG(FNCS|1, "Entering xli_2M_open()");
	file->private_data = vdev;

	down(&xli->lock);

	err = -ENODEV;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if (!xli->intf || !interface_to_usbdev(xli->intf) || !xli->present) {
#else
	if (!xli->dev || !xli->present) {
#endif
		err("intf=0x%p", xli->intf);
		if (xli->intf) err("dev=0x%p", interface_to_usbdev(xli->intf));
		err("xli->present=%d", xli->present);
		err("No camera plugged in");
		goto out;
	}
	
	if (xli->bridge == NULL || xli->bridge->sensor == CC_UNKNOWN) {
		err("No sensor is detected yet");
		goto out;
	}

	err = 0;
	
	if (xli->user == 0)
		err = ovfx2_initial_open(xli->bridge, file);
	
	if (err == 0)
		xli->user++;

out:
	up(&xli->lock);
	PDEBUG(FNCS|1, "Exiting xli_2M_open()");
	return err;
}


static int
xli_2M_ioctl(struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
	struct video_device *vdev = file->private_data;
	struct usb_xli_2M *xli = video_get_drvdata(vdev);

	if (!xli || !xli->present) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_ioctl(): no device");
		return -ENODEV;
	}

	return ovfx2_ioctl(xli->bridge, inode, file, cmd, arg);
}


static int
xli_2M_read(struct file *file, char *buf, size_t cnt, loff_t *ppos)
{
	struct video_device *vdev = file->private_data;
	struct usb_xli_2M *xli = video_get_drvdata(vdev);

	if (!xli || !xli->present) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_read(): no device");
		return -EFAULT;
	}

	return ovfx2_read(xli->bridge, file, buf, cnt, ppos);
}


static int
xli_2M_mmap(struct file *file, struct vm_area_struct *vma)
{
	struct video_device *vdev = file->private_data;
	struct usb_xli_2M *xli = video_get_drvdata(vdev);

	if (!xli || !xli->present) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_mmap(): fault");
		return -EFAULT;
	}

	return ovfx2_mmap(xli->bridge, file, vma);
}


/*
 * We come here as a result of a close() call from user-space.
 */
static int
xli_2M_close(struct inode *inode, struct file *file)
{
	struct video_device *vdev = file->private_data;
	struct usb_xli_2M *xli = video_get_drvdata(vdev);

	if (!xli) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_close(): no device");
		return -ENODEV;
	}

	PDEBUG(FNCS|1, "Entering xli_2M_close()");

	down(&xli->lock);
	
	if (xli->user > 0) {
		xli->user--;
# if 1
		if (xli->user == 0 && xli->present)
			/* last one out please put out the lights */
			ovfx2_final_close(xli->bridge);
# endif
	}
	
	up(&xli->lock);
# if 1
	down(&xli_free_lock);
	
	if (!xli->present) {
		PDEBUG(MISC|2, "close of already-unplugged device");
		
		if (xli->user == 0) {
			PDEBUG(MISC|1, "final close of already-unplugged device");
			video_set_drvdata(vdev, NULL); /* get the vdev struct to forget about the 'xli' struct */
			xli_2M_unregister(xli);		/* unregister V4L */
			ovfx2_free(xli->bridge);
			xli->bridge = NULL;
		
		//	kfree(xli);
		//	xli = NULL;
		}
	}
	
	up(&xli_free_lock);
# endif
	file->private_data = NULL;
	PDEBUG(FNCS|1, "Exiting xli_2M_close()");
	return 0;
}


static struct file_operations xli_2M_fops = {
	.owner =	THIS_MODULE,
	.open =		xli_2M_open,
	.release =	xli_2M_close,
	.read =		xli_2M_read,
	.mmap =		xli_2M_mmap,
	.ioctl =	xli_2M_ioctl,
	.llseek =	no_llseek,
};

static struct video_device vdev_template = {
	.owner =	THIS_MODULE,
	.name =		"2.1Mpixel USB Camera",
	.type =		VID_TYPE_CAPTURE,
	.hardware =	VID_HARDWARE_OV511,	/* FIXME */
	.fops =		&xli_2M_fops,
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	.release =	video_device_release,
#endif
	.minor =	-1,
};


/****************************************************************************
 *
 *  USB routines
 *
 ***************************************************************************/
/*
 * Omnivision provided a reference design and firmware connecting an FX2
 * to many of their I2C-aware sensor arrays.  Different products using
 * this design could have different sensors.
 */
static struct usb_device_id device_table [] = {
	/*
	 * Orange Micro "iBOT2" uses an ov7620.
	 *
	 * HIGH SPEED:
	 *
	 * T:  Bus=04 Lev=01 Prnt=01 Port=00 Cnt=01 Dev#=  2 Spd=480 MxCh= 0
	 * D:  Ver= 2.00 Cls=00(>ifc ) Sub=00 Prot=00 MxPS=64 #Cfgs=  1
	 * P:  Vendor=0b62 ProdID=0059 Rev= 1.00
	 * S:  Manufacturer=Orange Micro
	 * S:  Product=iBOT2 Camera
	 * C:* #Ifs= 1 Cfg#= 1 Atr=80 MxPwr=400mA
	 * I:  If#= 0 Alt= 0 #EPs= 2 Cls=ff(vend.) Sub=00 Prot=00 Driver=(none)
	 * E:  Ad=81(I) Atr=03(Int.) MxPS=   4 Ivl=  4ms
	 * E:  Ad=82(I) Atr=02(Bulk) MxPS= 512 Ivl=  1ms
	 *
	 * FULL SPEED:
	 *
	 * T:  Bus=02 Lev=01 Prnt=01 Port=00 Cnt=01 Dev#=  3 Spd=12  MxCh= 0
	 * D:  Ver= 2.00 Cls=00(>ifc ) Sub=00 Prot=00 MxPS=64 #Cfgs=  1
	 * P:  Vendor=0b62 ProdID=0059 Rev= 1.00
	 * S:  Manufacturer=Orange Micro
	 * S:  Product=iBOT2 Camera
	 * C:* #Ifs= 1 Cfg#= 1 Atr=80 MxPwr=400mA
	 * I:  If#= 0 Alt= 0 #EPs= 2 Cls=ff(vend.) Sub=00 Prot=00 Driver=(none)
	 * E:  Ad=81(I) Atr=03(Int.) MxPS=   4 Ivl=  1ms
	 * E:  Ad=82(I) Atr=02(Bulk) MxPS=  64 Ivl=  1ms
	 *
	 */
	{ USB_DEVICE (VEND_ORANGE_MICRO, PROD_IBOT2),
		.driver_info = (unsigned long) "Orange Micro iBOT2",
		},
	{ USB_DEVICE (VEND_OMNIVISION, PROD_TEST),
		.driver_info = (unsigned long) "OV PC Cam",
		},
	{ USB_DEVICE (VEND_XLI, PROD_XLI_2M),
		.driver_info = (unsigned long) "XLI 2.1 MegaPixel Camera",
		},

	{ }
};

MODULE_DEVICE_TABLE (usb, device_table);

/*
 * We come here as a result of either:
 * 1) Issuing an "insmod" on the driver with the camera plugged in at the time.
 * 2) The camera being plugged in whilst the driver is loaded.
 *
 * It seems not to be possible to tell the difference between these two
 * instances.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
static int
xli_2M_probe(struct usb_interface *intf, const struct usb_device_id *id)
{
	struct usb_device *dev = interface_to_usbdev(intf);
#else
static void *
xli_2M_probe(struct usb_device *dev, unsigned int ifnum,
	    const struct usb_device_id *id)
{
#endif
	struct usb_xli_2M *xli;
	int i;

	PDEBUG(FNCS|1, "Entering xli_2M_probe()");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	/* Since code below may sleep, we use this as a lock */
	MOD_INC_USE_COUNT;
#endif
	if (dev->speed != USB_SPEED_HIGH) {
		err("USB camera isn't plugged into a high speed port");
		goto error_out;
	}
	
	if ((xli = kmalloc(sizeof(*xli), GFP_KERNEL)) == NULL) {
		err("couldn't kmalloc xli struct");
		goto error_out;
	}

	memset(xli, 0, sizeof(*xli));
	PDEBUG(MISC|1, "xli struct allocated");

	init_MUTEX(&xli->lock);	/* to 1 == available */

	/* link ourselves to V4L */
	if ((xli->vdev = video_device_alloc()) == NULL)
		goto error_out;

	memcpy(xli->vdev, &vdev_template, sizeof(*xli->vdev));
	PDEBUG(MISC|3, "video device allocated");
	
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	xli->vdev->dev = &dev->dev;
#endif
	video_set_drvdata(xli->vdev, xli);	/* outside this driver, only vdev knows the address of this 'xli' struct */

	for (i = 0; i < XLI_2M_MAX_UNIT_VIDEO; i++) {
		/* Minor 0 cannot be specified; assume user wants autodetect */
		if (unit_video[i] == 0)
			break;

		if (video_register_device(xli->vdev, VFL_TYPE_GRABBER,
			unit_video[i]) >= 0) {
			break;
		}
	}

	/* Use the next available one */
	if ((xli->vdev->minor == -1) &&
	    video_register_device(xli->vdev, VFL_TYPE_GRABBER, -1) < 0) {
		err("video_register_device failed");
		goto error_out;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	usb_set_intfdata(intf, xli);
	xli->intf = intf;
//	xli_create_sysfs(xli->vdev);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	if ((xli->bridge = ovfx2_probe(intf, id, xli->vdev->minor)) == NULL) {
#else
	if ((xli->bridge = ovfx2_probe(dev, id, xli->vdev->minor)) == NULL) {
#endif
		/* failed to probe Omnivision's "FX2-9000" bridge chip? */
		err("couldn't find an FX2-9000 USB bridge");
		goto error_out;
	}
# if 0
	/*
	 * I (Steve Hosgood) have not yet done this much of the split.
	 * Currently, the sensor is probed by a side effect of ovfx2_probe() registering
	 * itself as a kernel i2c server.
	 *
	 * The prober for the sensor is compiled into ovcamchip.ko
	 *
	 * Most of the time the sensor is prodded, it's done directly by ovfx2.ko
	 * Quite a mess....
	 */
	if ((xli->sensor = ov2610_probe(ovfx2_get_i2c_interface(xli->bridge))) == NULL) {
		/* failed to probe Omnivision 2610 or 2610AE? */
		err("couldn't find an OV2610 sensor");
		ovfx2_disconnect(xli->bridge);
		goto error_out;
	}
	
	/* tell the bridge about the sensor since the bridge's state-machines need to talk to it */
	ovfx2_set_sensor(xli->bridge, xli->sensor);
# endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 20)
	if (usb_make_path(dev, xli->usb_path, XLI_USB_PATH_LEN) < 0) {
		err("usb_make_path error");
		goto error_out;
	}
#else
	snprintf(xli->usb_path, XLI_USB_PATH_LEN, "dev:%d/bus:%d",
		dev->bus->busnum, dev->devnum);
#endif

	info("%s at %s registered to video%d (%s speed)",
	     (char *) id->driver_info, xli->usb_path, xli->vdev->minor,
	     (dev->speed == USB_SPEED_HIGH) ? "high" : "full");

	PDEBUG(MISC|3, "registered as video%d", xli->vdev->minor);
	
	/* Update I2C adapter name with minor # */
	sprintf(i2c_adapname(&xli->bridge->i2c_adap), "xli_2M #%d", xli->vdev->minor);
	
	xli->present = 1;
	info("XLI 2M camera probe complete");
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	PDEBUG(FNCS|1, "Exiting XLI_2M_probe()");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
	return 0;
#else
	return xli;
#endif

error_out:
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	err("Camera initialization failed");
	PDEBUG(FNCS|1, "Exiting xli_2M_probe(): failed");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 36)
	return -EIO;
#else
	return NULL;
#endif
}


/*
 * We come here as a result of either:
 * 1) Issuing an "rmmod" on the driver with the camera plugged in at the time.
 * 2) The camera being unplugged whilst the driver is loaded.
 *
 * It seems not to be possible to tell the difference between these two
 * instances.
 */
static void
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 0)
xli_2M_disconnect(struct usb_interface *intf)
{
	struct usb_xli_2M *xli = usb_get_intfdata(intf);
#else
xli_2M_disconnect(struct usb_device *dev, void *ptr)
{
	struct usb_xli_2M *xli = (struct usb_xli_2M *) ptr;
#endif
	
	if (!xli) {
		PDEBUG(FNCS|1, "Repelled from xli_2M_disconnect(): no device");
		return;
	}
	
	PDEBUG(FNCS|1, "Entering xli_2M_disconnect()");

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_INC_USE_COUNT;
#endif

	/*
	 * We may have been in the middle of streaming.
	 * KLUDGE: can't take 'xli->lock' here because if we *are* streaming, user-code may well be
	 * waiting for a CSYNC in the IOCTL code (with xli->lock held), and we'll be deadlocked.
	 */
//	down(&xli->lock);
	xli->present = 0;
	ovfx2_disconnect(xli->bridge);
	
	/*
	 * We can take xli->lock now: ovfx2_disconnect will have released any active CSYNC call.
	 * No more will occur because xli->present is now zero.
	 */
	down(&xli->lock);
	
	if (xli->user == 0) {
		PDEBUG(MISC|1, "disconnect of already-closed device");
		xli_2M_unregister(xli);		/* unregister V4L */
		ovfx2_free(xli->bridge);
		xli->bridge = NULL;
		up(&xli->lock);	
		
	//	down(&xli_free_lock);
	//	kfree(xli);
	//	xli = NULL;
	//	up(&xli_free_lock);
	}
	else {
		PDEBUG(MISC|1, "disconnect of still-open device");
		up(&xli->lock);	
	}
	
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 20)
	MOD_DEC_USE_COUNT;
#endif
	PDEBUG(FNCS|1, "Exiting xli_2M_disconnect()");
	return;
}


static struct usb_driver xli_2M_driver = {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 4, 20)
	.owner =	THIS_MODULE,
#endif
	.name =		"xli_2M",
	.id_table =	device_table,
	.probe =	xli_2M_probe,
	.disconnect =	xli_2M_disconnect
};


/****************************************************************************
 *
 *  Module routines
 *
 ***************************************************************************/

/*
 * We come here as a consequence of loading the driver with '/sbin/insmod'
 */ 
static int __init
xli_2M_init(void)
{

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
	EXPORT_NO_SYMBOLS;
#endif
	init_MUTEX(&xli_free_lock);
	
# ifdef MODULE_DEBUG
	if ((debug & ~LMASK) == 0)
		/* in case user forgot to specify */
		debug |= MISC;
# endif /* MODULE_DEBUG */

	/* NB: string pasting below! */
	printk(KERN_INFO DESC " " VERSION "\n");
	printk(KERN_INFO "Written by: " AUTHOR "\n");
	printk(KERN_INFO "(c) " COPYRIGHT "\n");

	proc_xli_2M_create();
	
	if (usb_register(&xli_2M_driver) < 0)		/* does a call to xli_2M_probe() if the camera is plugged in right now */
		return -1;
		
	printk(KERN_INFO DESC " loaded OK\n");
	return 0;
}


/*
 * We come here as a consequence of unloading the driver with '/sbin/rmmod'
 */ 
static void __exit
xli_2M_cleanup(void)
{

	usb_deregister(&xli_2M_driver);			/* does a call to xli_2M_disconnect() if the camera is plugged in right now */
//	info("driver deregistered");

        proc_xli_2M_destroy();

	/* NB: string pasting below! */
	printk(KERN_INFO DESC " unloaded\n");
	return;
}


# ifdef MODULE
module_init(xli_2M_init);
module_exit(xli_2M_cleanup);
# endif
